"""
Note, scans and gonios not supported here. This writer essentially writes still images

Example to write 10 cbfs to a single NXmx file:
writer = NXmxWriter("*.cbf")
writer.write_nxmx("example.h5")
"""

from __future__ import annotations

import datetime
import os
import sys

import h5py
import numpy as np
from serialtbx.detector import basis

from cctbx import factor_ev_angstrom
from libtbx import easy_pickle
from libtbx.phil import parse
from libtbx.utils import Sorry
from scitbx import matrix
from scitbx.array_family import flex

from dials.util.options import ArgumentParser, flatten_experiments

from dxtbx import flumpy
from dxtbx.format.FormatCBFMultiTile import angle_and_axis

help_message = """
Create a NeXus file from either an experiment list or a set of image files

Uses NXmx as documented here:
https://manual.nexusformat.org/classes/applications/NXmx.html
and in Bernstein et. al. (2020):
https://doi.org/10.1107/S2052252520008672
"""

phil_scope = parse(
    """
  output_file = None
    .type = path
    .help = output file path
  wavelength = None
    .type = float
    .help = If not provided, use data from files provided.
  mask_file = None
    .type = str
    .help = Path to file with bad pixel mask in DIALS mask format
  trusted_range = None
    .type = floats(size=2)
    .help = Override the trusted range
  compression = gzip
    .type = str
    .help = Compression to apply to the data
  dtype = None
    .type = str
    .help = Override the data type. If data is floats and an integer type is specified, \
            the data is first rounded.
  nexus_details {
    instrument_name = None
      .type = str
      .help = Name of instrument. Consistency with the controlled vocabulary beamline naming in \
              https://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_diffrn_source.pdbx_synchrotron_beamline.html \
              and https://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_diffrn_source.type.html \
              is highly recommended.
    instrument_short_name = None
      .type = str
      .help = short name for instrument, perhaps the acronym
    source_name = None
      .type = str
      .help = Name of the neutron or x-ray storage ring/facility. Consistency with the naming in \
              https://mmcif.wwpdb.org/dictionaries/mmcif_pdbx_v50.dic/Items/_diffrn_source.pdbx_synchrotron_site.html \
              controlled vocabulary is highly recommended.
    source_short_name = None
      .type = str
      .help = short name for source, perhaps the acronym
    start_time = None
      .type = str
      .help = ISO 8601 time/date of the first data point collected in UTC, \
              using the Z suffix to avoid confusion with local time
    end_time = None
      .type = str
      .help = ISO 8601 time/date of the last data point collected in UTC, \
              using the Z suffix to avoid confusion with local time. \
              This field should only be filled when the value is accurately \
              observed. If the data collection aborts or otherwise prevents \
              accurate recording of the end_time, this field should be omitted
    end_time_estimated = None
      .type = str
      .help = ISO 8601 time/date of the last data point collected in UTC, \
              using the Z suffix to avoid confusion with local time. \
              This field may be filled with a value estimated before an \
              observed value is avilable.
    count_time = None
      .type = float
      .help = Elapsed actual counting time
    frame_time = None
      .type = float
      .help = This is time for each frame. This is exposure_time + readout time \
              in seconds.
    sample_name = None
      .type = str
      .help = Descriptive name of sample
    total_flux = None
      .type = float
      .help = flux incident on beam plane in photons per second
  }
"""
)

# Conversion from the imgCIF coordinate system conventionally used by dxtbx to
# the McStas coordinate system as used by NeXus:
#   https://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/Caxis.html
#   https://manual.nexusformat.org/design.html#design-coordinatesystem
IMGCIF_TO_MCSTAS = matrix.diag([-1, 1, -1])


class NXmxWriter:
    """Class for writing NXmx NeXus files from any dxtbx-supported format class"""

    def __init__(self, params, experiments=None, imageset=None):
        self.params = params
        if experiments or imageset:
            self.setup(experiments, imageset)
        self.handle = None

    def setup(self, experiments=None, imageset=None):
        assert [experiments, imageset].count(
            None
        ) == 1, "Supply either experiments or imagset, not both"
        if experiments:
            self.imagesets = experiments.imagesets()
            assert len(experiments.detectors()) == 1, "Multiple detectors not supported"
            self.detector = experiments.detectors()[0]
            self.beams = experiments.beams()
            assert len(experiments.scans()) <= 1
            assert len(experiments.goniometers()) <= 1
            self.scan = experiments[0].scan
            self.goniometer = experiments[0].goniometer

            if self.scan or self.goniometer:
                assert self.scan and self.goniometer
                assert len(experiments) == 1
        else:
            self.imagesets = [imageset]
            self.detector = imageset.get_detector(0)
            self.beams = [imageset.get_beam(i) for i in range(len(imageset))]

            self.scan = imageset.get_scan(0)
            self.goniometer = imageset.get_goniometer(0)
        self.construct_entry()

    def construct_entry(self):
        output_file_name = (
            self.params.output_file
            if self.params.output_file is not None
            else "converted.h5"
        )
        self.handle = h5py.File(output_file_name, "w")
        f = self.handle
        f.attrs["NX_class"] = "NXroot"
        f.attrs["file_name"] = os.path.basename(output_file_name)
        f.attrs["file_time"] = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")
        f.attrs["HDF5_Version"] = h5py.version.hdf5_version
        entry = f.create_group("entry")
        entry.attrs["NX_class"] = "NXentry"
        if self.params.nexus_details.start_time:
            entry["start_time"] = self.params.nexus_details.start_time
        if self.params.nexus_details.end_time:
            entry["end_time"] = self.params.nexus_details.end_time
        if self.params.nexus_details.end_time_estimated:
            entry["end_time_estimated"] = self.params.nexus_details.end_time_estimated

        # --> definition
        self._create_scalar(entry, "definition", "S4", np.string_("NXmx"))

        # --> sample
        sample = self.handle["entry"].create_group("sample")
        sample.attrs["NX_class"] = "NXsample"
        if self.params.nexus_details.sample_name:
            sample["name"] = self.params.nexus_details.sample_name
        sample["depends_on"] = "."  # Will be overriden if a scan and gonio are added

    def __call__(self, experiments=None, imageset=None):
        if experiments or imageset:
            self.setup(experiments, imageset)
        self.validate()
        self.construct_detector()
        self.add_all_beams()
        self.append_all_frames()
        self.add_scan_and_gonio()

    def validate(self):
        if not self.params.nexus_details.instrument_name:
            raise ValueError("instrument_name is required.")
        if not self.params.nexus_details.source_name:
            raise ValueError("source_name is required.")
        if not self.params.nexus_details.start_time:
            raise ValueError("start_time is required.")
        if not self.params.nexus_details.end_time_estimated:
            raise ValueError("end_time_estimated is required.")
        if not self.params.nexus_details.sample_name:
            raise ValueError("sample_name is required.")

    def get_metrology_dict(self):
        """Build a metrology dictionary.  This dictionary maps hierarchy keys to basis
        objects. A hierarchy key looks like this (0,1,2), where the entries are
        levels in a hierarchy and the numbers refer to a panel or group within that
        level"""
        metro = {}

        def recursive_setup_dict(panelgroup, key):
            metro[key] = basis(panelgroup=panelgroup)
            if panelgroup.is_panel():
                return
            for i, child in enumerate(panelgroup):
                childkey = tuple(list(key) + [i])
                recursive_setup_dict(child, childkey)

        recursive_setup_dict(self.detector.hierarchy(), (0,))
        return metro

    def _create_scalar(self, handle, path, dtype, value):
        dataset = handle.create_dataset(path, (), dtype=dtype)
        dataset[()] = value

    def create_vector(self, handle, name, value, **attributes):
        handle.create_dataset(name, (1,), data=[value], dtype="f")
        for key, attribute in attributes.items():
            handle[name].attrs[key] = attribute

    def add_frame_shift(self, handle, basis):
        """Add an axis representing a frame shift (a rotation axis with an offset)"""
        angle, axis = angle_and_axis(basis)

        if angle == 0:
            axis = matrix.col((0, 0, 1))

        if basis.include_translation:
            translation = basis.translation
        else:
            translation = matrix.col((0, 0, 0))

        axis = IMGCIF_TO_MCSTAS * axis
        translation = IMGCIF_TO_MCSTAS * translation

        self.create_vector(
            handle,
            os.path.basename(basis.name),
            angle,
            depends_on=basis.depends_on,
            equipment="detector",
            equipment_component=basis.equipment_component,
            transformation_type="rotation",
            units="deg",
            vector=axis,
            offset=translation,
            offset_units="mm",
        )

    def construct_detector(self, detector=None):
        """
        Hierarchical structure of master nexus file. Format information available here
        http://download.nexusformat.org/sphinx/classes/base_classes/NXdetector_module.html#nxdetector-module
        --> entry
          --> data
          --> definition (leaf)
          --> instrument
          --> sample
        """
        if not self.handle:
            # setup was skipped, so not using imagesets/experiments
            self.construct_entry()
            assert detector
            self.detector = detector
        if not detector:
            detector = self.detector

        # set up the metrology dictionary to include axis names, pixel sizes, and so forth
        metro = self.get_metrology_dict()

        def panel_group_from_key(key):
            # Find the node that a hierarchy key refers to
            if len(key) == 1:
                assert key[0] == 0
                return detector.hierarchy()

            node = detector.hierarchy()
            for i in key[1:]:
                node = node[i]
            return node

        def level_string(key):
            # Example for key (0,1,2). "L0M0_L1M1_L2M2", where L is level and M is module
            return "_".join(["L%dM%d" % (l, m) for l, m in enumerate(key)])

        def recursive_setup_basis_dict(key, parent_name="", panel_id=0):
            # Set up NeXus axis names, including equipment components and depends_on chains
            basis = metro[key]
            node = panel_group_from_key(key)
            nodename = level_string(key)
            if basis.name:
                dname = os.path.split(basis.name)[1]
            else:
                basis.name = dname = "AXIS_" + nodename

            if node.is_panel():
                panel_id += 1

            if len(key) == 1:
                assert key == (0,)  # only one detector allowed for now
                basis.depends_on = "."
            else:
                basis.depends_on = parent_name

            basis.equipment_component = "detector_level_%d" % (len(key) - 1)
            basis.axis_name = dname
            if not node.is_panel():
                for c, child in enumerate(node):
                    panel_id = recursive_setup_basis_dict(
                        tuple(list(key) + [c]), dname, panel_id
                    )
            return panel_id

        f = self.handle
        entry = f["entry"]
        recursive_setup_basis_dict((0,))

        # --> source
        source = entry.create_group("source")
        source.attrs["NX_class"] = "NXsource"
        source["name"] = self.params.nexus_details.source_name
        if self.params.nexus_details.source_short_name:
            source["name"].attrs[
                "short_name"
            ] = self.params.nexus_details.source_short_name
        # --> instrument
        instrument = entry.create_group("instrument")
        instrument.attrs["NX_class"] = "NXinstrument"
        instrument["name"] = self.params.nexus_details.instrument_name
        if self.params.nexus_details.instrument_short_name:
            instrument["name"].attrs[
                "short_name"
            ] = self.params.nexus_details.instrument_short_name
        beam = instrument.create_group("beam")
        beam.attrs["NX_class"] = "NXbeam"
        if self.params.nexus_details.total_flux:
            self._create_scalar(
                beam, "total_flux", "f", self.params.nexus_details.total_flux
            )
            beam["total_flux"].attrs["units"] = "Hz"
            beam.attrs["flux"] = "total_flux"

        det_group = instrument.create_group("detector_group")
        det_group.attrs["NX_class"] = "NXdetector_group"

        det_group.create_dataset("group_index", data=list(range(1, 3)), dtype="i")
        data = [np.string_("detector"), np.string_("detector")]
        det_group.create_dataset("group_names", (2,), data=data, dtype="S12")
        det_group.create_dataset("group_parent", (2,), data=[-1, 1], dtype="i")
        det_group.create_dataset("group_type", (2,), data=[1, 2], dtype="i")
        det = instrument.create_group("detector")
        det.attrs["NX_class"] = "NXdetector"
        det["description"] = "Detector converted from DIALS models"
        det["depends_on"] = "/entry/instrument/detector/transformations/AXIS_RAIL"
        det["gain_setting"] = "auto"
        assert len(set([p.get_material() for p in detector])) == 1
        assert len(set([p.get_thickness() for p in detector])) == 1
        det["sensor_material"] = detector[0].get_material()
        self._create_scalar(
            det, "sensor_thickness", "f", detector[0].get_thickness() * 1000
        )
        det["sensor_thickness"].attrs["units"] = "microns"
        if self.params.nexus_details.count_time is not None:
            self._create_scalar(
                det, "count_time", "f", self.params.nexus_details.count_time
            )
            det["count_time"].attrs["units"] = "us"
        if self.params.nexus_details.frame_time is not None:
            self._create_scalar(
                det, "frame_time", "f", self.params.nexus_details.frame_time
            )
            det["frame_time"].attrs["units"] = "s"

        if self.params.mask_file is not None:
            mask = easy_pickle.load(self.params.mask_file)
            if len(mask) > 1:
                mask = [flumpy.to_numpy(m) for m in mask]
                mask = np.stack(mask)
            else:
                mask = flumpy.to_numpy(mask[0])
            mask = (~mask).astype(np.int32)
            det.create_dataset("pixel_mask", mask.shape, data=mask, dtype=mask.dtype)

        transformations = det.create_group("transformations")
        transformations.attrs["NX_class"] = "NXtransformations"

        if self.params.trusted_range is None:
            assert len(set([p.get_trusted_range() for p in detector])) == 1
            trusted_min, trusted_max = detector[0].get_trusted_range()
        else:
            trusted_min, trusted_max = self.params.trusted_range
        # DIALS definitions match up with NXmx
        det.create_dataset("underload_value", (1,), data=[trusted_min], dtype="int32")
        det.create_dataset("saturation_value", (1,), data=[trusted_max], dtype="int32")

        def find_panel_id(panel):
            for i in range(len(detector)):
                if detector[i].get_name() == panel.get_name():
                    return i

        # Create a series of axis describing frame shifts from each level of the detector to the next
        for key in sorted(metro):
            basis = metro[key]
            node = panel_group_from_key(key)

            if node.is_panel():
                aname = level_string(key)
                fast = (
                    IMGCIF_TO_MCSTAS
                    * basis.orientation.unit_quaternion_as_r3_rotation_matrix()
                    * matrix.col((1.0, 0.0, 0.0))
                )
                slow = (
                    IMGCIF_TO_MCSTAS
                    * basis.orientation.unit_quaternion_as_r3_rotation_matrix()
                    * matrix.col((0.0, 1.0, 0.0))
                )

                module = det.create_group(aname)
                module.attrs["NX_class"] = "NXdetector_module"

                if len(detector) > 1:
                    panel_id = find_panel_id(node)
                    module.create_dataset(
                        "data_origin", (3,), data=[panel_id, 0, 0], dtype="i"
                    )
                    module.create_dataset(
                        "data_size",
                        (3,),
                        data=[1, *list(reversed(node.get_image_size()))],
                        dtype="i",
                    )
                else:
                    module.create_dataset("data_origin", (2,), data=[0, 0], dtype="i")
                    module.create_dataset(
                        "data_size",
                        (2,),
                        data=list(reversed(node.get_image_size())),
                        dtype="i",
                    )

                self.create_vector(
                    module,
                    "fast_pixel_direction",
                    node.get_pixel_size()[1],
                    depends_on=transformations.name
                    + "/"
                    + os.path.basename(basis.depends_on),
                    transformation_type="translation",
                    units="mm",
                    vector=fast,
                    offset=IMGCIF_TO_MCSTAS * basis.translation,
                    offset_units="mm",
                )
                self.create_vector(
                    module,
                    "slow_pixel_direction",
                    node.get_pixel_size()[0],
                    depends_on=transformations.name
                    + "/"
                    + os.path.basename(basis.depends_on),
                    transformation_type="translation",
                    units="mm",
                    vector=slow,
                    offset=IMGCIF_TO_MCSTAS * basis.translation,
                    offset_units="mm",
                )
            else:
                self.add_frame_shift(transformations, basis)

    def add_all_beams(self):
        spectra = []
        for imageset in self.imagesets:
            spectra.extend(
                [
                    imageset.get_spectrum(i)
                    for i in range(len(imageset))
                    if imageset.get_spectrum(i)
                ]
            )
        self.add_beams(self.beams, spectra)

    def add_beams(self, beams=None, spectra=None):
        handle = self.handle["entry/instrument/beam"]
        if self.params.wavelength:
            assert beams is None and spectra is None
            handle.create_dataset(
                "incident_wavelength", (1,), data=self.params.wavelength, dtype="f8"
            )
        else:
            assert beams
            if spectra:
                assert len(beams) == len(spectra)
                if len(beams) > 1:
                    test = None
                    for spectrum in spectra[1:]:
                        if test:
                            test &= (
                                spectrum.get_energies_eV()
                                == spectra[0].get_energies_eV()
                            )
                        else:
                            test = (
                                spectrum.get_energies_eV()
                                == spectra[0].get_energies_eV()
                            )
                    matching_channels = test.count(False) == 0

                    if matching_channels:
                        spectra_x = (
                            factor_ev_angstrom
                            / spectra[0].get_energies_eV().as_numpy_array()
                        )
                    else:
                        spectra_x = np.stack(
                            [
                                factor_ev_angstrom
                                / s.get_energies_eV().as_numpy_array()
                                for s in spectra
                            ]
                        )
                    spectra_y = np.stack(
                        [s.get_weights().as_numpy_array() for s in spectra]
                    )
                else:
                    spectra_x = (
                        factor_ev_angstrom
                        / spectra[0].get_energies_eV().as_numpy_array()
                    )
                    spectra_y = spectra[0].get_weights().as_numpy_array()

                matching_weighted_wavelengths = (
                    spectra[0].get_weighted_wavelength() == beams[0].get_wavelength()
                )
                if matching_weighted_wavelengths:
                    handle.create_dataset(
                        "incident_wavelength",
                        spectra_x.shape,
                        data=spectra_x,
                        dtype=spectra_x.dtype,
                    )
                    handle.create_dataset(
                        "incident_wavelength_weights",
                        spectra_y.shape,
                        data=spectra_y,
                        dtype=spectra_y.dtype,
                    )
                else:
                    wavelengths = np.array(
                        [beam.get_wavelength() for beam in self.beams],
                        dtype="f8",
                    )
                    handle.create_dataset(
                        "incident_wavelength",
                        wavelengths.shape,
                        data=wavelengths,
                        dtype=wavelengths.dtype,
                    )
                    handle.create_dataset(
                        "incident_wavelength_1Dspectrum",
                        spectra_x.shape,
                        data=spectra_x,
                        dtype=spectra_x.dtype,
                    )
                    handle.create_dataset(
                        "incident_wavelength_1Dspectrum_weights",
                        spectra_y.shape,
                        data=spectra_y,
                        dtype=spectra_y.dtype,
                    )
                    handle["incident_wavelength_1Dspectrum"].attrs["units"] = "angstrom"
                    handle["incident_wavelength"].attrs[
                        "variant"
                    ] = "incident_wavelength_1Dspectrum"
            else:
                if len(beams) > 1:
                    wavelengths = np.array(
                        [beam.get_wavelength() for beam in self.beams],
                        dtype="f8",
                    )
                    handle.create_dataset(
                        "incident_wavelength",
                        wavelengths.shape,
                        data=wavelengths,
                        dtype=wavelengths.dtype,
                    )
                else:
                    wavelength = self.beams[0].get_wavelength()
                    handle.create_dataset(
                        "incident_wavelength", (1,), data=wavelength, dtype="f8"
                    )
        handle["incident_wavelength"].attrs["units"] = "angstrom"

    def add_beam_in_sequence(self, beam, spectrum=None):
        """
        Alternate way to add beams, assuming adding them one at a time

        For spectra, this usecase assumes only one set of energy channels and no variants
        """
        handle = self.handle["entry/instrument/beam"]
        if spectrum:
            spectrum_x = (
                factor_ev_angstrom / spectrum.get_energies_eV().as_numpy_array()
            )
            spectrum_y = spectrum.get_weights().as_numpy_array()

            if "incident_wavelength" in handle:
                weights = handle["incident_wavelength_weights"]
                weights.resize(weights.shape[0] + 1, axis=0)
            else:
                handle.create_dataset(
                    "incident_wavelength",
                    spectrum_x.shape,
                    data=spectrum_x,
                    dtype=spectrum_x.dtype,
                )
                weights = handle.create_dataset(
                    "incident_wavelength_weights",
                    (1, len(spectrum_y)),
                    maxshape=(None, len(spectrum_y)),
                    dtype=spectrum_y.dtype,
                )
            weights[-1] = spectrum_y
        else:
            wavelength = beam.get_wavelength()
            if "incident_wavelength" in handle:
                dset = handle["incident_wavelength"]
                dset.resize(dset.shape[0] + 1, axis=0)
            else:
                dset = handle.create_dataset(
                    "incident_wavelength", (1,), maxshape=(None,), dtype="f8"
                )
            dset[-1] = wavelength
        handle["incident_wavelength"].attrs["units"] = "angstrom"

    def append_all_frames(self):
        """
        Given a h5py handle, append all the data. from the imagesets in the
        original experiment list.
        """
        for imageset in self.imagesets:
            for i in range(len(imageset)):
                self.append_frame(data=imageset[i])

    def append_frame(self, index=None, data=None):
        """
        Append some data. Can specify either
        the index into imageset zero, or the data itself.
        """
        if data is None:
            if index is None:
                data = self.imagesets[0][0]
            else:
                data = self.imagesets[0][index]
        if not isinstance(data, tuple):
            data = (data,)

        if len(data) > 1:
            assert len(set([d.focus() for d in data])) == 1
            shape = len(data), data[0].focus()[0], data[0].focus()[1]
        else:
            shape = data[0].focus()[0], data[0].focus()[1]

        dataisint = []
        for panel_data in data:
            assert len(panel_data.focus()) == 2
            if isinstance(panel_data, flex.int):
                dataisint.append(True)
            elif isinstance(panel_data, flex.double):
                dataisint.append(False)
            else:
                raise TypeError("Ints or doubles are required")
        assert all(dataisint) or not any(dataisint), "Mix of ints and doubles found"
        dataisint = all(dataisint)

        if self.params.dtype:
            if self.params.dtype in ["int", "float"]:
                dtype = int if self.params.dtype == "int" else float
            else:
                dtype = np.dtype(self.params.dtype)
            if not dataisint and np.issubdtype(dtype, np.integer):
                data = [p.iround() for p in data]
        else:
            dtype = int if dataisint else float

        det = self.handle["entry/instrument/detector"]
        entry = self.handle["entry"]
        if "data" in entry:
            data_group = entry["data"]
            dset = data_group["data"]
            dset.resize(dset.shape[0] + 1, axis=0)
        else:
            data_group = entry.create_group("data")
            data_group.attrs["NX_class"] = "NXdata"
            dset = data_group.create_dataset(
                "data",
                (1, *shape),
                maxshape=(None, *shape),
                dtype=dtype,
                compression=self.params.compression,
            )

        if len(data) > 1:
            data = [flumpy.to_numpy(d) for d in data]
            data = np.stack(data)
        else:
            data = flumpy.to_numpy(data[0])
        dset[-1:] = data.astype(dtype)

        if "bit_depth_readout" not in det:
            self._create_scalar(det, "bit_depth_readout", "i", dset.dtype.itemsize * 8)

    def add_scan_and_gonio(self, scan=None, gonio=None):
        if scan is None or gonio is None:
            assert scan is None and gonio is None
            scan = self.scan
            gonio = self.goniometer
        if scan is None or gonio is None:
            return

        sample = self.handle["entry/sample"]
        if "depends_on" in sample:
            del sample["depends_on"]
        multi_axis = hasattr(gonio, "get_axes")

        if multi_axis:
            root = gonio.get_names()[0]
        else:
            root = "omega"
        sample["depends_on"] = "/entry/sample/transformations/%s" % (root)

        transformations = sample.create_group("transformations")
        transformations.attrs["NX_class"] = "NXtransformations"

        def setup_axis(name, vector, main_axis=False):
            if main_axis:
                angles = np.array(
                    [
                        scan.get_angle_from_array_index(i)
                        for i in range(*scan.get_array_range())
                    ]
                )
            else:
                angles = [0]
            axis = transformations.create_dataset(name, data=angles)

            if main_axis:
                oscillation_angle = scan.get_oscillation()[1]
                oscillation = np.array([oscillation_angle] * len(scan))
                transformations.create_dataset(
                    name + "_increment_set", data=oscillation
                )

                end = angles + oscillation
                transformations.create_dataset(name + "_end", data=end)
            axis.attrs["transformation_type"] = "rotation"
            axis.attrs["units"] = "deg"
            axis.attrs["vector"] = IMGCIF_TO_MCSTAS.as_numpy_array() @ np.array(vector)
            return axis

        if multi_axis:
            for axis_number in range(len(gonio.get_axes())):
                if axis_number == gonio.get_scan_axis():
                    axis = setup_axis(
                        gonio.get_names()[axis_number],
                        gonio.get_axes()[axis_number],
                        main_axis=True,
                    )
                else:
                    axis = setup_axis(
                        gonio.get_names()[axis_number],
                        gonio.get_axes()[axis_number],
                        main_axis=False,
                    )

                if axis_number == len(gonio.get_axes()) - 1:
                    axis.attrs["depends_on"] = "."
                else:
                    axis.attrs["depends_on"] = "/entry/sample/transformations/%s" % (
                        gonio.get_names()[axis_number + 1]
                    )
        else:
            setup_axis("omega", gonio.get_rotation_axis(), main_axis=True)

        assert len(set(scan.get_exposure_times())) == 1
        if self.params.nexus_details.frame_time is None:
            self._create_scalar(
                self.handle["entry/instrument/detector"],
                "frame_time",
                "f",
                scan.get_exposure_times()[0],
            )
            self.handle["entry/instrument/detector/frame_time"].attrs["units"] = "s"

    def setup_multi_crystal(self, multi_crystal_group_name):
        self.handle["entry/sample"].create_group(multi_crystal_group_name)

    def add_crystal(self, multi_crystal_group_name, crystal):
        from dials.utils.nexus import nx_mx

        # Assume this is the next crystal in the list, added after the last image
        # was added
        n_images = self.handle["entry/data/data"].shape[0]
        nx_mx.dump_crystal(
            self.handle["entry/sample/" + multi_crystal_group_name],
            crystal,
            None,
            "crystal%d" % (n_images - 1),
        )


def run(args):
    usage = "dials.python nxmx_writer.py <experiments OR image files>"
    parser = ArgumentParser(
        usage=usage,
        sort_options=True,
        phil=phil_scope,
        read_experiments=True,
        read_experiments_from_images=True,
        epilog=help_message,
    )
    params, options = parser.parse_args(args=args, show_diff_phil=True)

    experiments = flatten_experiments(params.input.experiments)

    try:
        NXmxWriter(params)(experiments)
    except ValueError as e:
        print("Missing information. Run with -c to see full parameter description")
        raise Sorry(e)


if __name__ == "__main__":
    run(sys.argv[1:])



"""
Helper code to assist with reading TIFF file headers, which are by their
nature binary so we need to mess with things like byte swapping.
"""

from __future__ import annotations

import struct
import sys

from dxtbx.format.Format import Format

LITTLE_ENDIAN = 1234
BIG_ENDIAN = 4321


def _tiff_byte_order(four_bytes):
    """Determine the byte order for the file from the magic numbers at the
    very start of the file."""

    if b"II" == four_bytes[:2]:
        if struct.unpack("<H", four_bytes[2:])[0] == 42:
            return LITTLE_ENDIAN
    elif b"MM" == four_bytes[:2]:
        if struct.unpack(">H", four_bytes[2:])[0] == 42:
            return BIG_ENDIAN
    return False


def read_basic_tiff_header(filename):
    """Read the TIFF header (assuming for the moment a 4k header...) and
    return ... something."""

    # things we hope to learn from the vanilla TIFF header

    image_width = None
    image_height = None
    image_depth = None
    header_size = None

    with Format.open_file(filename, "rb") as fh:
        tiff_header = fh.read(1024)
    byte_order = _tiff_byte_order(tiff_header[:4])
    if not byte_order:
        return False

    if byte_order == LITTLE_ENDIAN:
        _I = "<I"
        _H = "<H"
    else:
        _I = ">I"
        _H = ">H"

    offset = struct.unpack(_I, tiff_header[4:8])[0]

    ntags = struct.unpack(_H, tiff_header[offset : offset + 2])[0]
    start = offset + 2

    for j in range(ntags):
        type_desc = struct.unpack(_H, tiff_header[start : start + 2])[0]
        start += 2
        type_type = struct.unpack(_H, tiff_header[start : start + 2])[0]
        start += 2
        # type_size = struct.unpack(_I, tiff_header[start : start + 4])[0]
        start += 4
        if type_type == 4:
            type_offset_or_value = struct.unpack(_I, tiff_header[start : start + 4])[0]
            start += 4
        elif type_type == 3:
            type_offset_or_value = struct.unpack(_H, tiff_header[start : start + 2])[0]
            start += 4

        if type_desc == 256:
            image_width = type_offset_or_value
        elif type_desc == 257:
            image_height = type_offset_or_value
        elif type_desc == 258:
            image_depth = type_offset_or_value
        elif type_desc == 273:
            header_size = type_offset_or_value

    return image_width, image_height, image_depth, header_size, byte_order


def _read_tiff_image_description(tiff_header, byte_order):
    """Search the TIFF header for an image description."""

    # OK then let's get started - and let's assume that the size is > 1 kb

    if byte_order == LITTLE_ENDIAN:
        _I = "<I"
        _H = "<H"
    else:
        _I = ">I"
        _H = ">H"

    offset = struct.unpack(_I, tiff_header[4:8])[0]
    ntags = struct.unpack(_H, tiff_header[offset : offset + 2])[0]
    start = offset + 2

    header_text = None

    for j in range(ntags):
        type_desc = struct.unpack(_H, tiff_header[start : start + 2])[0]
        start += 2
        type_type = struct.unpack(_H, tiff_header[start : start + 2])[0]
        start += 2
        type_size = struct.unpack(_I, tiff_header[start : start + 4])[0]
        start += 4
        if type_type == 4:
            type_offset_or_value = struct.unpack(_I, tiff_header[start : start + 4])[0]
            start += 4
        elif type_type == 3:
            type_offset_or_value = struct.unpack(_H, tiff_header[start : start + 2])[0]
            start += 4
        elif type_type == 2:
            type_offset_or_value = struct.unpack(_I, tiff_header[start : start + 4])[0]
            start += 4

        if type_desc == 270:
            start = type_offset_or_value
            end = type_offset_or_value + type_size
            header_text = tiff_header[start:end].strip()

    return header_text


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        width, height, depth, header, order = read_basic_tiff_header(arg)
        print("(%d x %d) @ %d + %d" % (width, height, depth, header))
        tiff_header = Format.open_file(arg, "rb").read(header)
        text = _read_tiff_image_description(tiff_header, order)
        if text:
            print(text)
        else:
            print("No text found")



from __future__ import annotations

import re

import h5py
import nxmx

from scitbx.array_family import flex

from dxtbx.format.FormatNXmx import FormatNXmx
from dxtbx.nexus import _dataset_as_flex, get_detector_module_slices

DATA_FILE_RE = re.compile(r"data_\d{6}")


class FormatNXmxEigerFilewriter(FormatNXmx):

    _cached_file_handle = None

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file) as handle:
            if "/entry/instrument/detector/detectorSpecific/eiger_fw_version" in handle:
                return True
        return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""
        super().__init__(image_file, **kwargs)

    def _start(self):
        super()._start()
        try:
            # This is (currently) a DECTRIS-specific non-standard item that
            # we will use in preference to bit_depth_readout (see below)
            self._bit_depth_image = int(
                self._cached_file_handle["/entry/instrument/detector/bit_depth_image"][
                    ()
                ]
            )
        except KeyError:
            self._bit_depth_image = None

    def _get_nxmx(self, fh: h5py.File):
        nxmx_obj = nxmx.NXmx(fh)
        nxentry = nxmx_obj.entries[0]

        nxdetector = nxentry.instruments[0].detectors[0]
        if nxdetector.underload_value is None:
            nxdetector.underload_value = 0

        # data_size is reversed - we should probably be more specific in when
        # we do this, i.e. check data_size is in a list of known reversed
        # values
        known_safe = [(1082,1035),]
        for module in nxdetector.modules:
            if not tuple(module.data_size) in known_safe:
                module.data_size = module.data_size[::-1]
        return nxmx_obj

    def get_raw_data(self, index):
        nxmx_obj = self._get_nxmx(self._cached_file_handle)
        nxdata = nxmx_obj.entries[0].data[0]
        nxdetector = nxmx_obj.entries[0].instruments[0].detectors[0]

        # Prefer bit_depth_image over bit_depth_readout since the former
        # actually corresponds to the bit depth of the images as stored on
        # disk. See also:
        #   https://www.dectris.com/support/downloads/header-docs/nexus/
        bit_depth = self._bit_depth_image or self._bit_depth_readout
        raw_data = get_raw_data(nxdata, nxdetector, index, bit_depth)

        if bit_depth:
            # if 32 bit then it is a signed int, I think if 8, 16 then it is
            # unsigned with the highest two values assigned as masking values
            if bit_depth == 32:
                top = 2**31
            else:
                top = 2**bit_depth
            for data in raw_data:
                d1d = data.as_1d()
                d1d.set_selected(d1d == top - 1, -1)
                d1d.set_selected(d1d == top - 2, -2)
        return raw_data


def get_raw_data(
    nxdata: nxmx.NXdata,
    nxdetector: nxmx.NXdetector,
    index: int,
    bit_depth: int | None = None,
) -> tuple[flex.float | flex.double | flex.int, ...]:
    """Return the raw data for an NXdetector.

    This will be a tuple of flex.float, flex.double or flex.int arrays, of length equal
    to the number of modules. The result is intended to be compatible with the
    get_raw_data() method of dxtbx format classes.

    Specialized version for files written by the DECTRIS EIGER filewriter, which don't
    write a virtual dataset, and instead we have to manually identify the correct
    data_nnnnnn sub-dataset for a given index.
    """
    data_subsets = [v for k, v in sorted(nxdata.items()) if DATA_FILE_RE.match(k)]
    for data in data_subsets:
        if index < data.shape[0]:
            break
        index -= data.shape[0]
    if index >= data.shape[0]:
        raise IndexError(f"Out of range index for raw data {index}")
    all_data = []
    sliced_outer = data[index]
    for module_slices in get_detector_module_slices(nxdetector):
        data_as_flex = _dataset_as_flex(
            sliced_outer, tuple(module_slices), bit_depth=bit_depth
        )
        all_data.append(data_as_flex)
    return tuple(all_data)



from __future__ import annotations

import itertools

import numpy as np
import pytest

import cctbx.array_family.flex as cctbx_flex
import scitbx.array_family.flex as flex

import dxtbx.flumpy as flumpy

# A mapping of which flex type should map to which numpy dtype - for verification
lookup_flex_type_to_numpy = {
    "uint8": "B",
    "uint16": "H",
    "uint32": "I",
    "size_t": "Q",
    "int8": "b",
    "int16": "h",
    "int": "i",
    "int32": "i",
    "int64": "q",
    "float": "f",
    "double": "d",
    "bool": "?",
    "complex_double": "D",
    "vec3_double": "d",
    "vec3_int": "i",
    "vec2_double": "d",
    "tiny_size_t_2": "Q",
    "miller_index": "i",
}

# On POSIX platforms, long and q are the same and numpy tends to give q's
# On Windows, defer to numpy long binding directly
if np.dtype("l").itemsize == np.dtype("q").itemsize:
    lookup_flex_type_to_numpy["long"] = "q"
else:
    lookup_flex_type_to_numpy["long"] = "l"


def test_basics():
    with pytest.raises(ValueError):
        s = flumpy.Scuffer(1)
        memoryview(s)

    i = flex.int(10)
    d = flex.double(10)

    flumpy.Scuffer(d)
    flumpy.Scuffer(i)


@pytest.fixture(
    params=[
        "size_t",
        "uint8",
        "uint16",
        "uint32",
        "int",
        "long",
        "int8",
        "int16",
        "int32",
        "int64",
        "float",
        "double",
    ]
)
def flex_numeric(request):
    flex_typename = request.param
    if not hasattr(flex, flex_typename):
        pytest.skip(f"Type flex.{flex_typename} not available on this flex instance")
    # Don't do long on windows, we prefer to get int
    if flex_typename == "long" and np.dtype("l").itemsize == np.dtype("i").itemsize:
        pytest.skip("On this platform, flex.long = flex.int so latter preferred")
    return getattr(flex, flex_typename)


def test_numeric_1d(flex_numeric):
    # 1d basics
    f1d = flex_numeric(range(10))
    as_np = flumpy.to_numpy(f1d)
    assert (f1d == as_np).all()
    # Change and make sure reflected in both
    as_np[2] = 42
    assert (f1d == as_np).all()
    assert as_np.dtype.char != "?"


def test_reverse_numeric_1d(flex_numeric):
    dtype = lookup_flex_type_to_numpy[flex_numeric.__name__]
    npo = np.array([240, 259, 144, 187]).astype(dtype)
    fo = flumpy.from_numpy(npo)
    assert isinstance(fo, flex_numeric)
    assert fo.all() == npo.shape
    assert all(fo[x] == npo[x] for x in range(4))
    npo[0] = 42
    assert all(fo[x] == npo[x] for x in range(4))
    assert fo[0] == 42


def test_numeric_2d(flex_numeric):
    grid = flex.grid(10, 10)
    fo = flex_numeric(grid)
    as_np = flumpy.to_numpy(fo)

    for i in range(10):
        for j in range(10):
            as_np[j, i] = i + j
            assert fo[j, i] == as_np[j, i]
            fo[j, i] = max(i, j) - min(i, j)
            assert fo[j, i] == as_np[j, i]


def test_reverse_numeric_2d(flex_numeric):
    dtype = lookup_flex_type_to_numpy[flex_numeric.__name__]
    npo = np.array(
        [[240, 259, 144, 187], [240, 259, 144, 187], [240, 259, 144, 187]]
    ).astype(dtype)
    fo = flumpy.from_numpy(npo)
    assert isinstance(fo, flex_numeric)
    assert fo.all() == npo.shape
    assert all(fo[x] == npo[x] for x in itertools.product(range(3), range(4)))
    npo[0] = 42
    assert all(fo[x] == npo[x] for x in itertools.product(range(3), range(4)))
    assert fo[0] == 42
    fo[0, 1] = 2
    assert npo[0, 1] == 2

    # Test zero-dimensional arrays
    npo_zero = np.zeros((0, 3))
    assert list(flumpy.from_numpy(npo_zero).all()) == [0, 3]
    assert flumpy.from_numpy(npo_zero).size() == 0


def test_numeric_4d(flex_numeric):
    # Check that we can think fourth-dimensionally
    grid = flex.grid(1, 9, 8, 5)
    fo = flex_numeric(grid)
    assert fo.nd() == 4
    as_np = flumpy.to_numpy(fo)
    for indices in itertools.product(range(1), range(9), range(8), range(5)):
        fo[indices] = sum(indices)
        assert fo[indices] == as_np[indices]


def test_reverse_numeric_4d(flex_numeric):
    dtype = lookup_flex_type_to_numpy[flex_numeric.__name__]
    npo = np.zeros((1, 9, 8, 5), dtype=dtype)
    fo = flumpy.from_numpy(npo)
    assert isinstance(fo, flex_numeric)
    assert fo.nd() == 4
    for indices in itertools.product(range(1), range(9), range(8), range(5)):
        fo[indices] = sum(indices)
        assert fo[indices] == npo[indices]


def test_bool():
    fo = flex.bool(flex.grid([5, 5, 5, 5, 5]))
    for indices in itertools.product(*([range(5)] * 5)):
        fo[indices] = sum(indices) % 3 == 0 or sum(indices) % 5 == 0
    as_np = flumpy.to_numpy(fo)
    for indices in itertools.product(*([range(5)] * 5)):
        assert fo[indices] == as_np[indices]
    assert fo.count(True) == as_np.sum()


def test_reverse_bool():
    sums = np.sum(np.indices((5, 5, 5, 5, 5)), axis=0)
    npo = np.logical_or(sums % 3 == 3, sums % 5 == 0)
    fo = flumpy.from_numpy(npo)
    for idx in itertools.product(*[range(5)] * 5):
        assert fo[idx] == npo[idx]
    assert fo.count(True) == npo.sum()


@pytest.mark.parametrize(
    "flex_vec", [flex.vec3_double, flex.vec3_int, cctbx_flex.miller_index]
)
def test_vec3(flex_vec):
    basic_vector = [(i, i * 2, i * 3) for i in range(10)]
    fo = flex_vec(basic_vector)
    as_np = flumpy.to_numpy(fo)
    assert (as_np == fo).all()
    as_np[0] = (0, 4, 0)
    as_np[1, 2] = 42
    assert fo[0] == (0, 4, 0)
    assert (as_np == fo).all()


@pytest.mark.parametrize(
    "flex_vec", [flex.vec3_double, flex.vec3_int, cctbx_flex.miller_index]
)
def test_reverse_vec3(flex_vec):
    dtype = lookup_flex_type_to_numpy[flex_vec.__name__]
    no = np.zeros((5, 3), dtype=dtype)
    fo = flumpy.vec_from_numpy(no)
    assert fo.all() == (5,)
    fo[0] = (1, 2, 3)
    assert (no[0] == (1, 2, 3)).all()

    with pytest.raises(ValueError):
        flumpy.vec_from_numpy(no.reshape((1, 15)))


@pytest.mark.parametrize("dtype", [np.int32, np.intc])
def test_reverse_vec3_dtype(dtype):
    no = np.zeros((5, 3), dtype=dtype)
    fo = flumpy.vec_from_numpy(no)
    assert fo.all() == (5,)
    fo[0] = (1, 2, 3)
    assert (no[0] == (1, 2, 3)).all()


@pytest.mark.parametrize("dtype", [np.int32, np.intc, int])
def test_reverse_miller_index(dtype):
    hkl = np.array([(1, 0, 0), (0, 1, 0), (0, 0, 1)], dtype=dtype)
    if dtype is int and np.dtype("l").itemsize != np.dtype("i").itemsize:
        with pytest.raises(ValueError):
            flumpy.miller_index_from_numpy(hkl)
    else:
        mi = flumpy.miller_index_from_numpy(hkl)
        assert isinstance(mi, cctbx_flex.miller_index)
        assert len(mi) == 3
        for i in range(len(mi)):
            assert (mi[i] == hkl[i]).all()

        with pytest.raises(ValueError):
            flumpy.miller_index_from_numpy(hkl.reshape((1, 9)))

        with pytest.raises(ValueError):
            flumpy.miller_index_from_numpy(hkl.astype(float))


@pytest.mark.parametrize("flex_vec", [flex.vec2_double, flex.tiny_size_t_2])
def test_vec2(flex_vec):
    basic_vector = [(i, i * 2) for i in range(10)]
    fo = flex_vec(basic_vector)
    as_np = flumpy.to_numpy(fo)
    assert (as_np == fo).all()
    as_np[0] = (0, 4)
    as_np[1, 1] = 42
    assert fo[0] == (0, 4)
    assert (as_np == fo).all()


@pytest.mark.parametrize("flex_vec", [flex.vec2_double, flex.tiny_size_t_2])
def test_reverse_vec2(flex_vec):
    dtype = lookup_flex_type_to_numpy[flex_vec.__name__]
    no = np.zeros((5, 2), dtype=dtype)
    fo = flumpy.vec_from_numpy(no)
    assert fo.all() == (5,)
    fo[0] = (1, 2)
    assert (no[0] == (1, 2)).all()


def test_mat3():
    fo = flex.mat3_double(10)
    as_np = flumpy.to_numpy(fo)
    for i in range(10):
        fo[i] = [1, i, 0, 0, 1, 0, i, 0, 1]
    assert (as_np.reshape(10, 9) == fo).all()


def test_reverse_mat3():
    # Want to test both N...x3x3 and N...x9
    no = np.zeros((10, 3, 3))
    for i in range(10):
        no[i] = [[1, i, 0], [0, 1, 0], [i, 0, 1]]
    fo = flumpy.mat3_from_numpy(no)
    assert (no.reshape(10, 9) == fo).all()

    flatter = np.copy(no.reshape(10, 9))
    flatter[0][0] = 14
    assert no[0][0][0] != 14
    fo_2 = flumpy.mat3_from_numpy(flatter)
    assert (flatter == fo_2).all()

    # Check we can't use unsupported dtypes
    with pytest.raises(ValueError):
        flumpy.mat3_from_numpy(np.zeros((10, 3, 3), dtype="i"))


def test_complex():
    fo = flex.complex_double(10)
    as_np = flumpy.to_numpy(fo)
    fo[1] = 3j
    fo[4] = 1.3 + 3j
    assert (as_np == fo).all()


def test_reverse_complex():
    npo = np.array([3j, 4j, 3 + 5j])
    fo = flumpy.from_numpy(npo)
    assert fo[2] == 3 + 5j
    assert (npo == fo).all()
    fo[0] = 1j
    assert npo[0] == 1j


def test_basic_fromnumpy():
    no = np.zeros(10, dtype="B")
    fo = flumpy.from_numpy(no)
    assert len(fo) == 10


def test_already_numpy():
    no = np.zeros(10, dtype="B")
    assert flumpy.to_numpy(no) is no


def test_already_flex():
    fo = flex.int(10)
    assert flumpy.from_numpy(fo) is fo


def test_flex_loop_nesting():
    fo = flex.int(10)
    npo = flumpy.to_numpy(fo)
    assert fo is flumpy.from_numpy(npo)

    # Now try vec
    fo = flex.complex_double(5)
    npo = flumpy.to_numpy(fo)
    assert flumpy.from_numpy(npo) is fo


@pytest.mark.xfail()
def test_flex_looping_vecs():
    # We want to try and avoid recursive nesting, but don't want to
    # return the original object if attempting to miscast vec<->numeric
    # however.... this means lots of conditions to check, so ignore now
    fo = flex.vec3_double(5)
    npo = flumpy.to_numpy(fo)
    assert flumpy.vec_from_numpy(npo) is fo

    # Don't be dumb when casting
    flex_nonvec = flex.double((9, 3))
    assert not flumpy.vec_from_numpy(flumpy.to_numpy(flex_nonvec)) is flex_nonvec

    # mat3
    fo = flex.mat3_double(5)
    npo = flumpy.to_numpy(fo)
    assert flumpy.mat3_from_numpy(npo) is fo


def test_numpy_loop_nesting():
    no = np.array([0, 1, 2])
    fo = flumpy.from_numpy(no)
    no_2 = flumpy.to_numpy(fo)
    assert no_2 is no


def test_noncontiguous():
    npo = np.zeros((10, 4, 3))
    flumpy.from_numpy(npo)

    with pytest.raises(ValueError):
        flumpy.from_numpy(npo[:, 1:])

    with pytest.raises(ValueError):
        flumpy.vec_from_numpy(npo[:, 1:])

    with pytest.raises(ValueError):
        flumpy.mat3_from_numpy(npo[:, 1:])

    # Test fortran order
    npo_f = np.zeros((10, 4, 3), order="F")

    with pytest.raises(ValueError):
        flumpy.from_numpy(npo_f)

    with pytest.raises(ValueError):
        flumpy.vec_from_numpy(npo_f)

    with pytest.raises(ValueError):
        flumpy.mat3_from_numpy(npo_f)


def test_nonowning():
    f_a = flex.double([0, 1, 2, 3, 4])
    n_b = flumpy.to_numpy(f_a)
    f_c = flumpy.from_numpy(n_b[1:])
    assert f_c[0] == 1
    f_c[1] = 9
    assert f_a[2] == 9
    assert n_b[2] == 9
    assert f_c[1] == 9


def test_int_long_degeneracy():
    if np.dtype("l").itemsize != np.dtype("i").itemsize:
        pytest.skip("Test only runs on platforms where int = long")
    npo = np.array([240, 259, 144, 187]).astype("l")
    fo = flumpy.from_numpy(npo)
    assert isinstance(fo, flex.int)

    assert fo.all() == npo.shape
    assert all(fo[x] == npo[x] for x in range(4))
    npo[0] = 42
    assert all(fo[x] == npo[x] for x in range(4))
    assert fo[0] == 42


def test_single_entry_vec_from_numpy():
    with pytest.raises(ValueError):
        flumpy.vec_from_numpy(np.array([2.0, 3.0, 4.0]))
    ao = flumpy.vec_from_numpy(np.array([[2.0, 3.0, 4.0]]))
    assert len(ao) == 1
    assert ao[0] == pytest.approx([2.0, 3.0, 4.0])



from __future__ import annotations

from cctbx import uctbx
from cctbx.eltbx import attenuation_coefficient
from cctbx.sgtbx import space_group, space_group_symbols
from iotbx.xds import xds_inp, xparm
from rstbx.cftbx.coordinate_frame_converter import coordinate_frame_converter
from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame
from scitbx import matrix

import dxtbx
from dxtbx.imageset import ImageSetFactory
from dxtbx.model import Crystal, MosaicCrystalKabsch2010, ParallaxCorrectedPxMmStrategy
from dxtbx.model.detector_helpers_types import detector_helpers_types


def to_imageset(input_filename, extra_filename=None):
    """Get an image set from the xds input filename plus an extra filename

    Params:
        input_filename The XDS.INP file
        extra_filename A (G)XPARM.XDS, INTGRATE.HKL or XDS_ASCII.HKL file

    Returns:
        The imageset

    """
    # Read the input filename
    handle = xds_inp.reader()
    handle.read_file(input_filename)

    # Get the template
    template = handle.name_template_of_data_frames[0].replace("?", "#")
    if template.endswith("h5"):
        template = template.replace("######", "master")
    image_range = handle.data_range
    detector_name = handle.detector

    if extra_filename is not None:
        # we can get all the extra dxtbx models from extra_filename
        check_format = False
    else:
        # we need the image files present to get the dxtbx models
        check_format = True

    # If an extra filename has been specified, try to load models
    if extra_filename:
        models = dxtbx.load(extra_filename)
        detector = models.get_detector()
        if detector_name.strip() in ("PILATUS", "EIGER") or handle.silicon is not None:
            if handle.silicon is None:
                table = attenuation_coefficient.get_table("Si")
                wavelength = models.get_beam().get_wavelength()
                mu = table.mu_at_angstrom(wavelength) / 10.0
            else:
                mu = handle.silicon
            t0 = handle.sensor_thickness
            for panel in detector:
                panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
                panel.set_trusted_range(
                    (handle.minimum_valid_pixel_value, handle.overload)
                )
        beam = models.get_beam()
        detector = models.get_detector()
        goniometer = models.get_goniometer()
        scan = models.get_scan()
        scan.set_image_range(image_range)
    else:
        beam = None
        detector = None
        goniometer = None
        scan = None

    # Create the imageset
    imageset = ImageSetFactory.from_template(
        template,
        image_range=image_range,
        check_format=check_format,
        beam=beam,
        detector=detector,
        goniometer=goniometer,
        scan=scan,
    )[0]

    return imageset


def to_crystal(filename):
    """Get the crystal model from the xparm file

    Params:
        filename The xparm/or integrate filename

    Return:
        The crystal model

    """
    # Get the real space coordinate frame
    cfc = coordinate_frame_converter(filename)
    real_space_a = cfc.get("real_space_a")
    real_space_b = cfc.get("real_space_b")
    real_space_c = cfc.get("real_space_c")
    sg = cfc.get("space_group_number")
    crystal_space_group = space_group(space_group_symbols(sg).hall())
    mosaicity = cfc.get("mosaicity")

    # Return the crystal model
    if mosaicity is None:
        crystal = Crystal(
            real_space_a=real_space_a,
            real_space_b=real_space_b,
            real_space_c=real_space_c,
            space_group=crystal_space_group,
        )
    else:
        crystal = MosaicCrystalKabsch2010(
            real_space_a=real_space_a,
            real_space_b=real_space_b,
            real_space_c=real_space_c,
            space_group=crystal_space_group,
        )
        crystal.set_mosaicity(mosaicity)
    return crystal


def xds_detector_name(dxtbx_name):
    """Translate from a xia2 name from the detector library to an XDS detector
    name."""
    # http://xds.mpimf-heidelberg.mpg.de/html_doc/xds_parameters.html#DETECTOR=

    if "pilatus" in dxtbx_name:
        return "PILATUS"
    if "eiger" in dxtbx_name:
        return "EIGER"
    if "rayonix" in dxtbx_name:
        return "CCDCHESS"
    if "adsc" in dxtbx_name:
        return "ADSC"
    if "holton" in dxtbx_name:
        return "ADSC"
    if "saturn" in dxtbx_name:
        return "SATURN"
    if "raxis" in dxtbx_name:
        return "RAXIS"
    if "mar-345" in dxtbx_name:
        return "MAR345"
    if "mar" in dxtbx_name:
        return "MAR"
    if "unknown" in dxtbx_name:
        return "ADSC"

    raise RuntimeError("detector %s unknown" % dxtbx_name)


class to_xds:
    """A class to export contents of a Sequence as XDS.INP or XPARM.XDS."""

    def __init__(self, sequence):
        self._sequence = sequence

        # detector dimensions in pixels
        self.detector_size = [
            int(
                max(
                    panel.get_raw_image_offset()[0] + panel.get_image_size()[0]
                    for panel in self.get_detector()
                )
            ),
            int(
                max(
                    panel.get_raw_image_offset()[1] + panel.get_image_size()[1]
                    for panel in self.get_detector()
                )
            ),
        ]
        self.fast, self.slow = self.detector_size

        if len(self.get_detector()) > 1:
            fast = self.get_detector()[0].get_parent_fast_axis()
            slow = self.get_detector()[0].get_parent_slow_axis()
            Rd = align_reference_frame(fast, (1, 0, 0), slow, (0, 1, 0))
            origin = Rd * matrix.col(self.get_detector()[0].get_parent_origin())
        else:
            fast = self.get_detector()[0].get_fast_axis()
            slow = self.get_detector()[0].get_slow_axis()
            Rd = align_reference_frame(fast, (1, 0, 0), slow, (0, 1, 0))
            origin = Rd * matrix.col(self.get_detector()[0].get_origin())

        self.detector_x_axis = (Rd * matrix.col(fast)).elems
        self.detector_y_axis = (Rd * matrix.col(slow)).elems

        F = Rd * matrix.col(fast)
        S = Rd * matrix.col(slow)
        N = F.cross(S)
        self.detector_normal = N.elems

        # assume all panels same pixel size
        self.pixel_size = self.get_detector()[0].get_pixel_size()

        centre = -(origin - origin.dot(N) * N)
        x = centre.dot(F)
        y = centre.dot(S)

        f, s = self.pixel_size
        self.detector_distance = origin.dot(N)
        # Need to add 0.5 because XDS seems to do centroids in fortran coords
        self.detector_origin = (x / f + 0.5, y / f + 0.5)

        self.imagecif_to_xds_transformation_matrix = Rd

        self.panel_limits = []
        self.panel_x_axis = []
        self.panel_y_axis = []
        self.panel_origin = []
        self.panel_distance = []
        self.panel_normal = []

        for panel_id, panel in enumerate(self.get_detector()):

            f = Rd * matrix.col(panel.get_fast_axis())
            s = Rd * matrix.col(panel.get_slow_axis())
            n = f.cross(s)

            xmin, ymin = panel.get_raw_image_offset()
            xmax = xmin + panel.get_image_size()[0]
            ymax = ymin + panel.get_image_size()[1]
            self.panel_limits.append((xmin + 1, xmax, ymin + 1, ymax))

            o = Rd * matrix.col(panel.get_origin())
            op = o.dot(n) * n
            d0 = matrix.col((-x, -y, self.detector_distance))
            orgsx = (op - o + d0).dot(f) / self.pixel_size[0] + xmin
            orgsy = (op - o + d0).dot(s) / self.pixel_size[1] + ymin
            panel_distance = op.dot(n) - d0.dot(n)

            # axes in local (i.e. detector) frame
            fl = matrix.col(panel.get_local_fast_axis())
            sl = matrix.col(panel.get_local_slow_axis())
            nl = fl.cross(sl)

            self.panel_x_axis.append(fl.elems)
            self.panel_y_axis.append(sl.elems)
            self.panel_normal.append(nl.elems)
            self.panel_origin.append((orgsx, orgsy))
            self.panel_distance.append(panel_distance)

        # Beam stuff
        self.wavelength = self.get_beam().get_wavelength()
        self.beam_vector = Rd * matrix.col(
            self.get_beam().get_sample_to_source_direction()
        )
        # just to make sure it is the correct length
        self.beam_vector = self.beam_vector.normalize()  # / self.wavelength
        self.beam_vector = (-self.beam_vector).elems

        # Scan and goniometer stuff
        self.starting_frame = self.get_scan().get_image_range()[0]
        self.starting_angle = self.get_scan().get_oscillation()[0]
        self.oscillation_range = self.get_scan().get_oscillation()[1]
        self.rotation_axis = (
            Rd * matrix.col(self.get_goniometer().get_rotation_axis())
        ).elems

    def get_detector(self):
        return self._sequence.get_detector()

    def get_goniometer(self):
        return self._sequence.get_goniometer()

    def get_beam(self):
        return self._sequence.get_beam()

    def get_scan(self):
        return self._sequence.get_scan()

    def get_template(self):
        try:
            return self._sequence.get_template()
        except AttributeError:
            return "FIXME####.h5"

    def XDS_INP(
        self,
        space_group_number=None,
        real_space_a=None,
        real_space_b=None,
        real_space_c=None,
        job_card="XYCORR INIT COLSPOT IDXREF DEFPIX INTEGRATE CORRECT",
    ):
        result = []

        assert [real_space_a, real_space_b, real_space_c].count(None) in (0, 3)

        sensor = self.get_detector()[0].get_type()
        fast, slow = self.detector_size
        f, s = self.pixel_size
        df = int(1000 * f)
        ds = int(1000 * s)

        # FIXME probably need to rotate by pi about the X axis

        detector = xds_detector_name(
            detector_helpers_types.get(sensor, fast, slow, df, ds)
        )
        trusted = self.get_detector()[0].get_trusted_range()

        # XDS OVERLOAD parameter is the maximum trusted value, not the first
        # overloaded value (https://xds.mr.mpg.de/html_doc/xds_parameters.html#OVERLOAD=)
        result.append(
            "DETECTOR=%s MINIMUM_VALID_PIXEL_VALUE=%d OVERLOAD=%d"
            % (detector, trusted[0], trusted[1])
        )

        if detector in ("PILATUS", "EIGER"):
            result.append(
                "SENSOR_THICKNESS= %.3f" % self.get_detector()[0].get_thickness()
            )
            if self.get_detector()[0].get_material():
                material = self.get_detector()[0].get_material()
                table = attenuation_coefficient.get_table(material)
                mu = table.mu_at_angstrom(self.wavelength) / 10.0
                result.append(
                    "!SENSOR_MATERIAL / THICKNESS %s %.3f"
                    % (material, self.get_detector()[0].get_thickness())
                )
                result.append("!SILICON= %f" % mu)

        result.append(
            "DIRECTION_OF_DETECTOR_X-AXIS= %.5f %.5f %.5f" % self.detector_x_axis
        )

        result.append(
            "DIRECTION_OF_DETECTOR_Y-AXIS= %.5f %.5f %.5f" % self.detector_y_axis
        )

        result.append("NX=%d NY=%d QX=%.4f QY=%.4f" % (fast, slow, f, s))

        result.append("DETECTOR_DISTANCE= %.6f" % self.detector_distance)
        result.append("ORGX= %.2f ORGY= %.2f" % self.detector_origin)
        result.append("ROTATION_AXIS= %.5f %.5f %.5f" % self.rotation_axis)
        result.append("STARTING_ANGLE= %.3f" % self.starting_angle)
        result.append("OSCILLATION_RANGE= %.3f" % self.oscillation_range)
        result.append("X-RAY_WAVELENGTH= %.5f" % self.wavelength)
        result.append(
            "INCIDENT_BEAM_DIRECTION= %.3f %.3f %.3f"
            % tuple([b / self.wavelength for b in self.beam_vector])
        )

        # FIXME LATER
        if hasattr(self.get_beam(), "get_polarization_fraction"):
            result.append(
                "FRACTION_OF_POLARIZATION= %.3f"
                % self.get_beam().get_polarization_fraction()
            )
            result.append(
                "POLARIZATION_PLANE_NORMAL= %.3f %.3f %.3f"
                % self.get_beam().get_polarization_normal()
            )
        template = self.get_template()
        if template.endswith("master.h5"):
            template = template.replace("master", "??????")
        result.append("NAME_TEMPLATE_OF_DATA_FRAMES= %s" % template.replace("#", "?"))
        result.append("TRUSTED_REGION= 0.0 1.41")

        for panel, (x0, _, y0, _) in zip(self.get_detector(), self.panel_limits):
            for f0, s0, f1, s1 in panel.get_mask():
                result.append(
                    "UNTRUSTED_RECTANGLE= %d %d %d %d"
                    % (f0 + x0 - 1, f1 + x0, s0 + y0 - 1, s1 + y0)
                )

        start_end = self.get_scan().get_image_range()

        if start_end[0] == 0:
            start_end = (1, start_end[1])

        result.append("DATA_RANGE= %d %d" % start_end)
        result.append("JOB=%s" % job_card)
        if space_group_number is not None:
            result.append("SPACE_GROUP_NUMBER= %i" % space_group_number)
        if [real_space_a, real_space_b, real_space_c].count(None) == 0:
            R = self.imagecif_to_xds_transformation_matrix
            unit_cell_a_axis = R * matrix.col(real_space_a)
            unit_cell_b_axis = R * matrix.col(real_space_b)
            unit_cell_c_axis = R * matrix.col(real_space_c)
            result.append("UNIT_CELL_A-AXIS= %.6f %.6f %.6f" % unit_cell_a_axis.elems)
            result.append("UNIT_CELL_B-AXIS= %.6f %.6f %.6f" % unit_cell_b_axis.elems)
            result.append("UNIT_CELL_C-AXIS= %.6f %.6f %.6f" % unit_cell_c_axis.elems)

        if len(self.panel_x_axis) > 1:
            for panel_id, panel_x_axis in enumerate(self.panel_x_axis):

                result.append("")
                result.append("!")
                result.append("! SEGMENT %d" % (panel_id + 1))
                result.append("!")
                result.append("SEGMENT= %d %d %d %d" % self.panel_limits[panel_id])
                result.append(
                    "DIRECTION_OF_SEGMENT_X-AXIS= %.5f %.5f %.5f" % panel_x_axis
                )

                result.append(
                    "DIRECTION_OF_SEGMENT_Y-AXIS= %.5f %.5f %.5f"
                    % self.panel_y_axis[panel_id]
                )

                result.append("SEGMENT_DISTANCE= %.3f" % self.panel_distance[panel_id])

                result.append(
                    "SEGMENT_ORGX= %.2f SEGMENT_ORGY= %.2f"
                    % self.panel_origin[panel_id]
                )
        return "\n".join(result)

    def xparm_xds(self, real_space_a, real_space_b, real_space_c, space_group):
        R = self.imagecif_to_xds_transformation_matrix
        unit_cell_a_axis = R * matrix.col(real_space_a)
        unit_cell_b_axis = R * matrix.col(real_space_b)
        unit_cell_c_axis = R * matrix.col(real_space_c)
        A_inv = matrix.sqr(
            unit_cell_a_axis.elems + unit_cell_b_axis.elems + unit_cell_c_axis.elems
        )
        metrical_matrix = (A_inv * A_inv.transpose()).as_sym_mat3()
        unit_cell = uctbx.unit_cell(metrical_matrix=metrical_matrix)

        return xparm.write(
            self.starting_frame,
            self.starting_angle,
            self.oscillation_range,
            self.rotation_axis,
            self.wavelength,
            self.beam_vector,
            space_group,
            unit_cell.parameters(),
            unit_cell_a_axis.elems,
            unit_cell_b_axis.elems,
            unit_cell_c_axis.elems,
            None,  # num_segments
            self.detector_size,
            self.pixel_size,
            self.detector_origin,
            self.detector_distance,
            self.detector_x_axis,
            self.detector_y_axis,
            self.detector_normal,
            segments=None,
            orientation=None,
        )



from __future__ import annotations

from iotbx.detectors.hamamatsu import HamamatsuImage

from dxtbx.format.FormatSMVADSC import FormatSMVADSC


class FormatSMVHamamatsu(FormatSMVADSC):
    @staticmethod
    def understand(image_file):

        size, header = FormatSMVHamamatsu.get_smv_header(image_file)

        wanted_header_items = ["DETECTOR_NAME"]
        if any(item not in header for item in wanted_header_items):
            return False

        return "hamamatsu" in header["DETECTOR_NAME"].lower()

    def detectorbase_start(self):
        self.detectorbase = HamamatsuImage(self._image_file)
        self.detectorbase.open_file = self.open_file
        self.detectorbase.readHeader()



from __future__ import annotations

import sys

import h5py

from dxtbx.format import nexus
from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.format.FormatStill import FormatStill


class FormatNexus(FormatHDF5):
    @staticmethod
    def understand(image_file):
        try:
            return nexus.is_nexus_file(image_file)
        except OSError:
            return False

    def _start(self):

        # Read the file structure
        self._reader = reader = nexus.NXmxReader(self._image_file)

        # Only support 1 set of models at the moment
        assert len(reader.entries) == 1, "Currently only supports 1 NXmx entry"
        assert len(reader.entries[0].data) == 1, "Currently only supports 1 NXdata"
        assert (
            len(reader.entries[0].instruments) == 1
        ), "Currently only supports 1 NXinstrument"
        assert len(reader.entries[0].samples) == 1, "Currently only supports 1 NXsample"
        assert (
            len(reader.entries[0].samples[0].beams) == 1
            or len(reader.entries[0].instruments[0].beams) == 1
        ), "Currently only supports 1 NXbeam"

        # Get the NXmx model objects
        entry = reader.entries[0]
        self.instrument = instrument = entry.instruments[0]
        detector = instrument.detectors[0]
        sample = entry.samples[0]
        beam = sample.beams[0] if sample.beams else instrument.beams[0]
        data = entry.data[0]

        # Construct the models
        self._beam_factory = nexus.BeamFactory(beam)
        self._beam_factory.load_model(0)

        self._setup_gonio_and_scan(sample, detector)

        if self._scan_model:
            array_range = self._scan_model.get_array_range()
            num_images = array_range[1] - array_range[0]
        else:
            num_images = 0

        if len(instrument.detector_groups) == 0:
            assert (
                len(reader.entries[0].instruments[0].detectors) == 1
            ), "Currently only supports 1 NXdetector unless in a detector group"
            assert (
                len(reader.entries[0].instruments[0].detectors[0].modules) == 1
            ), "Currently only supports 1 NXdetector_module unless in a detector group"

            self._raw_data = nexus.DataFactory(data, max_size=num_images)
            self._detector_model = nexus.DetectorFactory(
                detector, self._beam_factory.model, shape=self._raw_data.shape()
            ).model
        else:
            self._raw_data = nexus.detectorgroupdatafactory(data, instrument)
            self._detector_model = nexus.DetectorFactoryFromGroup(
                instrument, self._beam_factory.model
            ).model

    def _setup_gonio_and_scan(self, sample, detector):
        """Set up rotation-specific models"""
        self._goniometer_model = nexus.GoniometerFactory(sample).model
        self._scan_model = nexus.generate_scan_model(sample, detector)

    def _end(self):
        return

    def _goniometer(self):
        return self._goniometer_model

    def _detector(self):
        return self._detector_model

    def _beam(self, index=None):
        self._beam_model, _ = self._beam_factory.read_models(index)
        return self._beam_model

    def _scan(self):
        return self._scan_model

    def get_goniometer(self, index=None):
        return self._goniometer()

    def get_detector(self, index=None):
        return self._detector()

    def get_beam(self, index=None):
        return self._beam(index)

    def get_spectrum(self, index=None):
        self._beam_model, _ = self._beam_factory.read_models(index)
        return self._beam_factory.spectrum

    def get_scan(self, index=None):
        if index is None:
            return self._scan()
        scan = self._scan()
        if scan is not None:
            return scan[index]
        return scan

    def get_raw_data(self, index):
        return self._raw_data[index]

    def get_static_mask(self, index=None, goniometer=None):
        return nexus.MaskFactory(self.instrument.detectors, index).mask

    def get_num_images(self):
        if self._scan() is not None:
            return self._scan().get_num_images()
        return len(self._raw_data)

    def get_image_file(self, index=None):
        return self._image_file

    def get_detectorbase(self, index=None):
        raise NotImplementedError

    @staticmethod
    def get_instrument_name(handle) -> str | None:
        if "short_name" in handle["/entry/instrument"].attrs:
            name = handle["/entry/instrument"].attrs["short_name"]
        elif "/entry/instrument/name" in handle:
            if "short_name" in handle["/entry/instrument/name"].attrs:
                name = handle["/entry/instrument/name"].attrs["short_name"]
            else:
                name = handle["/entry/instrument/name"][()]
        else:
            name = None
        return nexus.h5str(name)


class FormatNexusStill(FormatNexus, FormatStill):
    def __init__(self, *args, **kwargs):
        self.lazy = kwargs.get("lazy", True)
        FormatNexus.__init__(self, *args, **kwargs)
        FormatStill.__init__(self, *args, **kwargs)

    @staticmethod
    def understand(image_file):
        is_nexus_still = False
        try:
            # Get the file handle
            with h5py.File(image_file, "r") as handle:
                if "/entry/sample/goniometer/omega_increment" in handle:
                    return False

                for entry in nexus.find_entries(handle):
                    for sample in nexus.find_class(entry, "NXsample"):
                        if "depends_on" not in sample:
                            is_nexus_still = True
        except OSError:
            return False
        return is_nexus_still

    def _setup_gonio_and_scan(self, sample, detector):
        """No rotation-specific models for stills"""
        self._goniometer_model = None
        self._scan_model = None

    def get_num_images(self):
        return len(self._raw_data)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        if FormatNexus.understand(arg):

            format_instance = FormatNexus(arg)

            beam = format_instance.get_beam()
            detector = format_instance.get_detector()
            goniometer = format_instance.get_goniometer()
            scan = format_instance.get_scan()

            iset = FormatNexus.get_imageset(arg)
            print(beam)
            print(detector)
            print(goniometer)
            print(scan)

            print(len(iset))



from __future__ import annotations

import sys

from iotbx.detectors.adsc import ADSCImage

from dxtbx.format.FormatSMVADSC import FormatSMVADSC


class FormatSMVADSCSN(FormatSMVADSC):
    """A class for reading SMV format ADSC images, with detector serial number"""

    @staticmethod
    def understand(image_file):
        # The header must include a DATE and an integer-valued DETECTOR_SN
        # for this format to apply.

        size, header = FormatSMVADSC.get_smv_header(image_file)

        if "DATE" not in header:
            return False
        try:
            int(header["DETECTOR_SN"])
        except (KeyError, ValueError):
            return False

        return True

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""

        # Mapping of serial numbers to models for known detectors
        self._sn_to_model = {
            401: "Q4U",
            402: "Q4",
            414: "Q4",
            423: "Q4R",
            428: "Q4R",
            429: "Q4",  # or Q4R?
            441: "Q210",
            442: "Q210",
            443: "Q210",
            444: "Q210",  # or Q210R?
            445: "Q210",
            446: "Q210",
            447: "Q210",
            448: "Q210",
            457: "Q210R",
            471: "Q270",
            472: "Q270",
            474: "Q270",
            901: "Q210",
            905: "Q315",
            907: "Q315R",  # or Q315?
            913: "Q315",
            917: "Q315R",
            923: "Q315R",
            925: "Q315",
            926: "Q315R",
            928: "Q315R",
            931: "Q315R",
            933: "Q315R",
        }

        super().__init__(image_file, **kwargs)

    def _adsc_module_gain(self, model=None):
        """Overload to look the model number up from the serial number table"""

        if model is None:
            sn = int(self._header_dictionary["DETECTOR_SN"])
            model = self._sn_to_model.get(sn)
        return super()._adsc_module_gain(model=model)

    def detectorbase_start(self):

        self.detectorbase = ADSCImage(self._image_file)
        self.detectorbase.open_file = self.open_file
        self.detectorbase.readHeader()


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVADSCSN.understand(arg))



"""Helpers for FormatCBFMiniPilatus..."""


from __future__ import annotations

import calendar
import time


def get_pilatus_timestamp(timestamp_string):
    if "." in timestamp_string:
        timestamp, milliseconds = timestamp_string.split(".")
    else:
        timestamp = timestamp_string
        milliseconds = "000"

    for format in ["%Y-%b-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S", "%Y/%b/%d %H:%M:%S"]:

        try:
            struct_time = time.strptime(timestamp, format)
            return calendar.timegm(struct_time) + float("0." + milliseconds)

        except Exception:
            pass

    raise RuntimeError("timestamp %s not recognised" % timestamp)



from __future__ import annotations

import math
from typing import Tuple

import pycbf

import libtbx.phil

try:
    from ..dxtbx_model_ext import Beam, PolychromaticBeam, Probe
except ModuleNotFoundError:
    from dxtbx_model_ext import Beam, PolychromaticBeam, Probe  # type: ignore

Vec3Float = Tuple[float, float, float]

beam_phil_scope = libtbx.phil.parse(
    """
  beam
    .expert_level = 1
    .short_caption = "Beam overrides"
  {
    type = *monochromatic polychromatic
      .type = choice
      .help = "Override the beam type"
      .short_caption = "beam_type"

    probe = *x-ray electron neutron
      .type = choice
      .help = "Override the beam probe"
      .short_caption = "beam_probe"

    wavelength = None
      .type = float
      .help = "Override the beam wavelength"

    direction = None
      .type = floats(size=3)
      .help = "Override the sample to source direction"
      .short_caption = "Sample to source direction"

    divergence = None
        .type = float
        .help = "Override the beam divergence"

    sigma_divergence = None
        .type = float
        .help = "Override the beam sigma divergence"

    polarization_normal = None
      .type = floats(size=3)
      .help = "Override the polarization normal"
      .short_caption = "Polarization normal"

    polarization_fraction = None
      .type = float(value_min=0.0, value_max=1.0)
      .help = "Override the polarization fraction"
      .short_caption = "Polarization fraction"

    transmission = None
        .type = float
        .help = "Override the transmission"
        .short_caption = "transmission"

    flux = None
        .type = float
        .help = "Override the flux"
        .short_caption = "flux"

    sample_to_source_distance = None
        .type = float
        .help = "Override the distance between sample and source (mm)"
  }
"""
)


class BeamFactory:
    """A factory class for beam objects, which encapsulate standard beam
    models. In cases where a full cbf description is available this
    will be used, otherwise simplified descriptions can be applied."""

    @staticmethod
    def from_phil(
        params: libtbx.phil.scope_extract,
        reference: Beam | PolychromaticBeam | None = None,
    ) -> Beam | PolychromaticBeam:
        """
        Convert the phil parameters into a beam model
        """

        # Check the input
        if reference is None:
            beam = (
                PolychromaticBeam() if params.beam.type == "polychromatic" else Beam()
            )
        else:
            beam = reference

        # Set the parameters
        if params.beam.type == "monochromatic":
            if params.beam.wavelength is not None:
                beam.set_wavelength(params.beam.wavelength)
            elif reference is None:
                raise RuntimeError("No wavelength set")
        if params.beam.direction is not None:
            beam.set_direction(params.beam.direction)
        elif reference is None:
            raise RuntimeError("No beam direction set")

        if params.beam.divergence is not None:
            beam.set_divergence(params.beam.divergence)
        if params.beam.sigma_divergence is not None:
            beam.set_sigma_divergence(params.beam.sigma_divergence)
        if params.beam.polarization_normal is not None:
            beam.set_polarization_normal(params.beam.polarization_normal)
        if params.beam.polarization_fraction is not None:
            beam.set_polarization_fraction(params.beam.polarization_fraction)
        if params.beam.transmission is not None:
            beam.set_transmission(params.beam.transmission)
        if params.beam.flux is not None:
            beam.set_flux(params.beam.flux)
        if params.beam.sample_to_source_distance is not None:
            beam.set_sample_to_source_distance(params.beam.sample_to_source_distance)
        if params.beam.probe != "x-ray":
            beam.set_probe(Beam.get_probe_from_name(params.beam.probe))

        return beam

    @staticmethod
    def from_dict(dict: dict, template: dict = None) -> Beam | PolychromaticBeam:
        """Convert the dictionary to a beam model"""

        if template is not None:
            if "__id__" in dict and "__id__" in template:
                assert (
                    dict["__id__"] == template["__id__"]
                ), "Beam and template dictionaries are not the same type."

        if dict is None and template is None:
            return None
        joint = template.copy() if template else {}
        joint.update(dict)

        # Create the model from the joint dictionary
        if "probe" not in joint:
            joint["probe"] = "x-ray"
        if joint.get("__id__") == "polychromatic":
            return PolychromaticBeam.from_dict(joint)

        return Beam.from_dict(joint)

    @staticmethod
    def make_beam(
        sample_to_source: Vec3Float = None,
        wavelength: float = None,
        s0: Vec3Float = None,
        unit_s0: Vec3Float = None,
        divergence: float = None,
        sigma_divergence: float = None,
    ) -> Beam:
        if divergence is None or sigma_divergence is None:
            divergence = 0.0
            sigma_divergence = 0.0

        if sample_to_source:
            assert wavelength
            return Beam(
                tuple(map(float, sample_to_source)),
                float(wavelength),
                float(divergence),
                float(sigma_divergence),
            )
        elif unit_s0:
            assert wavelength
            return Beam(
                tuple(-float(x) for x in unit_s0),
                float(wavelength),
                float(divergence),
                float(sigma_divergence),
            )
        else:
            assert s0
            return Beam(tuple(map(float, s0)))

    @staticmethod
    def make_polychromatic_beam(
        direction: Vec3Float,
        divergence: float = 0.0,
        sigma_divergence: float = 0.0,
        polarization_normal: Vec3Float = (0.0, 1.0, 0.0),
        polarization_fraction: float = 0.5,
        flux: float = 0.0,
        transmission: float = 1.0,
        probe: Probe = Probe.xray,
        sample_to_source_distance: float = 0.0,
        deg: bool = True,
    ) -> PolychromaticBeam:
        return PolychromaticBeam(
            tuple(map(float, direction)),
            float(divergence),
            float(sigma_divergence),
            tuple(map(float, polarization_normal)),
            float(polarization_fraction),
            float(flux),
            float(transmission),
            probe,
            float(sample_to_source_distance),
            bool(deg),
        )

    @staticmethod
    def make_polarized_beam(
        sample_to_source: Vec3Float = None,
        wavelength: float = None,
        s0: Vec3Float = None,
        unit_s0: Vec3Float = None,
        polarization: Vec3Float = None,
        polarization_fraction: float = None,
        divergence: float = None,
        sigma_divergence: float = None,
        flux: float = None,
        transmission: float = None,
        probe: Probe = Probe.xray,
    ) -> Beam:
        assert polarization
        assert 0.0 <= polarization_fraction <= 1.0

        if divergence is None or sigma_divergence is None:
            divergence = 0.0
            sigma_divergence = 0.0

        if flux is None:
            flux = 0
        if transmission is None:
            transmission = 1.0

        if sample_to_source:
            assert wavelength
            return Beam(
                tuple(map(float, sample_to_source)),
                float(wavelength),
                float(divergence),
                float(sigma_divergence),
                tuple(map(float, polarization)),
                float(polarization_fraction),
                float(flux),
                float(transmission),
                probe,
            )
        elif unit_s0:
            assert wavelength
            return Beam(
                tuple(-float(x) for x in unit_s0),
                float(wavelength),
                float(divergence),
                float(sigma_divergence),
                tuple(map(float, polarization)),
                float(polarization_fraction),
                float(flux),
                float(transmission),
                probe,
            )
        else:
            assert s0
            sum_sq_s0 = s0[0] ** 2 + s0[1] ** 2 + s0[2] ** 2
            assert sum_sq_s0 > 0
            wavelength = 1.0 / math.sqrt(sum_sq_s0)
            return Beam(
                tuple(map(float, s0)),
                wavelength,
                float(divergence),
                float(sigma_divergence),
                tuple(map(float, polarization)),
                float(polarization_fraction),
                float(flux),
                float(transmission),
                probe,
            )

    @staticmethod
    def simple(wavelength: float) -> Beam:
        """Construct a beam object on the principle that the beam is aligned
        with the +z axis, as is quite normal. Also assume the beam has
        polarization fraction 0.999 and is polarized in the x-z plane, unless
        it has a wavelength shorter than 0.05 Å in which case we assume
        electron diffraction and return an unpolarized beam model."""

        if wavelength > 0.05:
            return BeamFactory.make_beam(
                sample_to_source=(0.0, 0.0, 1.0), wavelength=wavelength
            )
        else:
            return BeamFactory.make_polarized_beam(
                sample_to_source=(0.0, 0.0, 1.0),
                wavelength=wavelength,
                polarization=(0, 1, 0),
                polarization_fraction=0.5,
            )

    @staticmethod
    def simple_directional(sample_to_source: Vec3Float, wavelength: float) -> Beam:
        """Construct a beam with direction and wavelength."""

        if wavelength > 0.05:
            return BeamFactory.make_beam(
                sample_to_source=sample_to_source, wavelength=wavelength
            )
        else:
            return BeamFactory.make_polarized_beam(
                sample_to_source=sample_to_source,
                wavelength=wavelength,
                polarization=(0, 1, 0),
                polarization_fraction=0.5,
            )

    @staticmethod
    def complex(
        sample_to_source: Vec3Float,
        polarization_fraction: float,
        polarization_plane_normal: Vec3Float,
        wavelength: float,
    ) -> Beam:
        """Full access to the constructor for cases where we do know everything
        that we need..."""

        return BeamFactory.make_polarized_beam(
            sample_to_source=sample_to_source,
            wavelength=wavelength,
            polarization=polarization_plane_normal,
            polarization_fraction=polarization_fraction,
        )

    @staticmethod
    def imgCIF(cif_file: str) -> Beam:
        """Initialize a detector model from an imgCIF file. N.B. the
        definition of the polarization plane is not completely helpful
        in this - it is the angle between the polarization plane and the
        +Y laboratory frame vector."""

        cbf_handle = pycbf.cbf_handle_struct()
        cbf_handle.read_widefile(cif_file.encode(), pycbf.MSG_DIGEST)

        result = BeamFactory.imgCIF_H(cbf_handle)

        return result

    @staticmethod
    def imgCIF_H(cbf_handle: pycbf.cbf_handle_struct) -> Beam:
        """Initialize a detector model from an imgCIF file. N.B. the
        definition of the polarization plane is not completely helpful
        in this - it is the angle between the polarization plane and the
        +Y laboratory frame vector. This example works from a cbf_handle,
        which is already configured."""

        d2r = math.pi / 180.0

        cbf_handle.find_category(b"axis")

        # find record with equipment = source

        try:
            cbf_handle.find_column(b"equipment")
            cbf_handle.find_row(b"source")

            # then get the vector and offset from this
            direction = []

            for j in range(3):
                cbf_handle.find_column(b"vector[%d]" % (j + 1))
                direction.append(cbf_handle.get_doublevalue())
        except Exception as e:
            if str(e).split()[-1] != "CBF_NOTFOUND":
                raise
            direction = [0, 0, 1]

        # and the wavelength
        wavelength = cbf_handle.get_wavelength()

        # and information about the polarization - FIXME this should probably
        # be a rotation about the beam not about the Z axis. Should also check
        # to see if this is Cu K-alpha wavelength (i.e. lab source...)

        try:
            polar_fraction, polar_angle = cbf_handle.get_polarization()
        except Exception:
            polar_fraction = 0.999
            polar_angle = 0.0

        polar_plane_normal = (
            math.sin(polar_angle * d2r),
            math.cos(polar_angle * d2r),
            0.0,
        )

        # and the flux if available
        try:
            cbf_handle.find_category(b"diffrn_radiation")
            cbf_handle.find_column(b"beam_flux")
            flux = cbf_handle.get_value()
        except Exception as e:
            if str(e).split()[-1] != "CBF_NOTFOUND":
                raise
            flux = None

        return BeamFactory.make_polarized_beam(
            sample_to_source=direction,
            wavelength=wavelength,
            polarization=polar_plane_normal,
            polarization_fraction=polar_fraction,
            flux=flux,
        )



from __future__ import annotations

import argparse

import dxtbx.util
from dxtbx import load


def overload(image_file):
    i = load(image_file)
    data = i.get_raw_data()
    if not isinstance(data, tuple):
        data = (data,)
    detector = i.get_detector()
    for pid, (d, p) in enumerate(zip(data, detector)):
        if max(d) > p.get_trusted_range()[1]:
            return True
    return False


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser()
    parser.add_argument("image_file", nargs="+")
    options = parser.parse_args(args)
    for image_file in options.image_file:
        if overload(image_file):
            print(image_file)


if __name__ == "__main__":
    run()



"""
An implementation of the SMV image reader for pseudo "ADSC" images, converted
from Pilatus images using iotbx debug_write.
"""

from __future__ import annotations

import sys

from iotbx.detectors import SMVImage

from dxtbx.format.FormatSMVADSC import FormatSMVADSC


class FormatSMVADSCDBG(FormatSMVADSC):
    """Format class for reading images converted from pilatus to smv adsc"""

    @staticmethod
    def understand(image_file):
        """Check to see if this has the correct dimensions of a pilatus"""

        size, header = FormatSMVADSC.get_smv_header(image_file)

        if int(header["SIZE1"]) == 2463 and int(header["SIZE2"]) == 2527:
            return True

        return False

    def _start(self):
        # read the headers, then swap size1 and size2
        FormatSMVADSC._start(self)
        self._header_dictionary["SIZE1"] = "2527"
        self._header_dictionary["SIZE2"] = "2463"

    def detectorbase_start(self):
        self.detectorbase = SMVImage(self._image_file)
        self.detectorbase.open_file = self.open_file
        self.detectorbase.readHeader()
        self.detectorbase.parameters["SIZE1"] = 2527
        self.detectorbase.parameters["SIZE2"] = 2463

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])
        beam_x = float(self._header_dictionary["BEAM_CENTER_X"])
        beam_y = float(self._header_dictionary["BEAM_CENTER_Y"])
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        # size1 and size2 swapped here
        image_size = (
            float(self._header_dictionary["SIZE2"]),
            float(self._header_dictionary["SIZE1"]),
        )
        max_trusted_value = 65535
        min_trusted_value = 0

        return self._detector_factory.simple(
            "CCD",
            distance,
            (beam_y, beam_x),
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            (min_trusted_value, max_trusted_value),
            [],
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVADSCDBG.understand(arg))



"""
Testing URI-passing in the Format class hierarchy
"""

from __future__ import annotations

from typing import Type
from unittest.mock import Mock

import pytest

import dxtbx
import dxtbx.format.Registry as Registry
import dxtbx.imageset
from dxtbx.format.Format import Format
from dxtbx.model.experiment_list import ExperimentListFactory


@pytest.fixture
def registry(monkeypatch):
    """Temporarily register a format class"""
    _temporary_formats = {}
    _original_get_format_class_for = Registry.get_format_class_for

    def _get_format_class_for(format_class_name: str) -> Type[Format]:
        """Intercept fetching format classes"""
        if format_class_name in _temporary_formats:
            return _temporary_formats[format_class_name]
        else:
            return _original_get_format_class_for(format_class_name)

    monkeypatch.setattr(Registry, "get_format_class_for", _get_format_class_for)

    class _registry:
        @staticmethod
        def register(format_class: Format) -> None:
            name = format_class.__name__
            _temporary_formats[name] = format_class
            monkeypatch.setitem(Registry._format_dag, name, [])
            _format_entry = list(Registry._format_dag["Format"]) + [name]
            monkeypatch.setitem(Registry._format_dag, "Format", _format_entry)

    return _registry


def test_registry_inject(registry, tmp_path):
    """meta-test of registry injection"""
    _hit_understand = []

    class TestClass(Format):
        @classmethod
        def understand(cls, filename):
            _hit_understand.append(True)
            return "TEST_FILE" in str(filename)

    filename_notest = tmp_path / "A_FILE"
    filename_notest.touch()
    filename = tmp_path / "TEST_FILE"
    filename.touch()
    assert not Registry.get_format_class_for_file(str(filename_notest))
    registry.register(TestClass)
    assert Registry.get_format_class_for("TestClass") is TestClass
    assert Registry.get_format_class_for_file(str(filename))
    assert _hit_understand == [True]


def test_custom_scheme_handler(registry):
    class SchemeHandler(Format):
        schemes = ["scheme"]

        @classmethod
        def understand(cls, endpoint):
            assert endpoint.startswith("scheme://")
            return True

    assert not Registry.get_format_class_for_file("scheme://something")
    registry.register(SchemeHandler)
    assert Registry.get_format_class_for_file("scheme://something") is SchemeHandler
    # Check dxtbx.load
    instance = dxtbx.load("scheme://something")
    assert isinstance(instance, SchemeHandler)


def test_multiple_scheme_handler(registry, tmp_path):
    _hits = []

    class SchemeHandler(Format):
        schemes = ["scheme", ""]

        @classmethod
        def understand(cls, endpoint):
            _hits.append(endpoint)
            return True

    registry.register(SchemeHandler)
    assert Registry.get_format_class_for_file("scheme://something") is SchemeHandler
    filename = tmp_path / "TEST_FILE"
    filename.touch()
    assert Registry.get_format_class_for_file(str(filename)) is SchemeHandler
    assert _hits == ["scheme://something", str(filename)]


def test_no_abs_via_scheme(registry, monkeypatch):
    # Tries to construct these and relies on pickleable
    monkeypatch.setattr(dxtbx.imageset, "ImageSetData", Mock())
    monkeypatch.setattr(dxtbx.imageset, "ImageSet", Mock())

    _hits = []

    class SufficientError(Exception):
        """Stop when we've reached the test point"""

    class SchemeHandler(Format):
        schemes = ["scheme"]

        @classmethod
        def understand(cls, endpoint):
            return True

        def _start(self):
            assert self._image_file == "scheme://something"
            _hits.append(self._image_file)
            # The second reconstruction call had the issue
            if len(_hits) == 2:
                raise SufficientError

    registry.register(SchemeHandler)
    with pytest.raises(SufficientError):
        # We don't want to try doing everything here, just check construction
        ExperimentListFactory.from_filenames(["scheme://something"])

    assert _hits == ["scheme://something", "scheme://something"]


def test_no_abs_load_dict(registry):
    class SufficientError(Exception):
        """Stop when we've reached the test point"""

    class SchemeHandler(Format):
        schemes = ["scheme"]

        @classmethod
        def understand(cls, endpoint):
            return True

        def _start(self):
            assert self._image_file == "scheme://something"
            raise SufficientError

    registry.register(SchemeHandler)

    with pytest.raises(SufficientError):
        ExperimentListFactory.from_dict(
            {
                "__id__": "ExperimentList",
                "experiment": [
                    {
                        "__id__": "Experiment",
                        "imageset": 0,
                    }
                ],
                "imageset": [
                    {
                        "__id__": "ImageSet",
                        "images": ["scheme://something"],
                    }
                ],
            }
        )



"""
An implementation of the SMV image reader for Rigaku Pilatus images.
Inherits from FormatSMVRigaku.
"""

from __future__ import annotations

import sys
import time

from boost_adaptbx.boost.python import streambuf
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.ext import read_int32
from dxtbx.format.FormatSMVRigaku import FormatSMVRigaku


class FormatSMVRigakuPilatus(FormatSMVRigaku):
    """A class for reading SMV format Rigaku Pilatus images."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like a Rigaku d*TREK SMVRigaku format image,
        i.e. we can make sense of it. Essentially that will be if it contains
        all of the keys we are looking for."""

        size, header = FormatSMVRigaku.get_smv_header(image_file)

        if "DETECTOR_TYPE" not in header:
            return False

        if header["DETECTOR_TYPE"] not in ["Pilatus 200K", "Pilatus 300K"]:
            return False

        return True

    def _goniometer(self):
        """Initialize the structure for the goniometer - this will need to
        correctly compose the axes given in the image header. In this case
        this is made rather straightforward as the image header has the
        calculated rotation axis stored in it. We could work from the
        rest of the header and construct a goniometer model."""

        axis = tuple(map(float, self._header_dictionary["ROTATION_VECTOR"].split()))

        return self._goniometer_factory.known_axis(axis)

    def _detector(self):
        """Return a model for the detector, allowing for two-theta offsets
        and the detector position. This will be rather more complex..."""

        detector_name = self._header_dictionary["DETECTOR_NAMES"].split()[0].strip()

        detector_axes = self.get_detector_axes(detector_name)
        fast = matrix.col(tuple(detector_axes[:3]))
        slow = matrix.col(tuple(detector_axes[3:]))
        distortion = self.get_distortion(detector_name)

        # multiply through by the distortion to get the true detector fast, slow

        detector_fast, detector_slow = (
            distortion[0] * fast + distortion[1] * slow,
            distortion[2] * fast + distortion[3] * slow,
        )

        beam_pixels = self.get_beam_pixels(detector_name)
        pixel_size = self.get_pixel_size(detector_name)
        image_size = self.get_image_size(detector_name)

        detector_origin = -(
            beam_pixels[0] * pixel_size[0] * detector_fast
            + beam_pixels[1] * pixel_size[1] * detector_slow
        )

        gonio_axes = self.get_gonio_axes(detector_name)
        gonio_values = self.get_gonio_values(detector_name)
        gonio_units = self._header_dictionary["%sGONIO_UNITS" % detector_name].split()
        gonio_num_axes = int(
            self._header_dictionary["%sGONIO_NUM_VALUES" % detector_name]
        )

        rotations = []
        translations = []

        for j, unit in enumerate(gonio_units):
            axis = matrix.col(gonio_axes[3 * j : 3 * (j + 1)])
            if unit == "deg":
                rotations.append(
                    axis.axis_and_angle_as_r3_rotation_matrix(gonio_values[j], deg=True)
                )
                translations.append(matrix.col((0.0, 0.0, 0.0)))
            elif unit == "mm":
                rotations.append(
                    matrix.sqr((1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0))
                )
                translations.append(gonio_values[j] * axis)
            else:
                raise RuntimeError("unknown axis unit %s" % unit)

        rotations.reverse()
        translations.reverse()

        for j in range(gonio_num_axes):
            detector_fast = rotations[j] * detector_fast
            detector_slow = rotations[j] * detector_slow
            detector_origin = rotations[j] * detector_origin
            detector_origin = translations[j] + detector_origin

        max_trusted_value = int(float(self._header_dictionary["SATURATED_VALUE"]))
        min_trusted_value = 0

        return self._detector_factory.complex(
            "PAD",
            detector_origin.elems,
            detector_fast.elems,
            detector_slow.elems,
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
        )

    def _beam(self):
        """Return a simple model for the beam."""

        beam_direction = self.get_beam_direction()
        p_fraction, p_plane = self.get_beam_polarization()

        wavelength = float(self._header_dictionary["SCAN_WAVELENGTH"])

        return self._beam_factory.complex(
            beam_direction, p_fraction, p_plane, wavelength
        )

    def _scan(self):
        """Return the scan information for this image."""
        epoch = time.strptime(
            self._header_dictionary["DTREK_DATE_TIME"], "%d-%b-%Y %H:%M:%S"
        )
        return self._create_single_SVM_scan(epoch)

    def get_raw_data(self):
        """Read the data - assuming it is streaming 4-byte unsigned ints following the
        header..."""

        assert len(self.get_detector()) == 1
        size = self.get_detector()[0].get_image_size()
        f = self.open_file(self._image_file)
        f.read(int(self._header_dictionary["HEADER_BYTES"]))
        raw_data = read_int32(streambuf(f), int(size[0] * size[1]))
        raw_data.reshape(flex.grid(size[1], size[0]))

        return raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVRigakuPilatus.understand(arg))



from __future__ import annotations

from pathlib import Path

import setuptools

from build import build

# Static version number which is updated by bump2version
# Do not change this manually - use 'bump2version <major/minor/patch/release>'
__version_tag__ = "3.17.0"

setup_kwargs = {
    "name": "dxtbx",
    "version": __version_tag__,
    "long_description": Path(__file__).parent.joinpath("README.md").read_text(),
    "description": "Diffraction Experiment Toolbox",
    "author": "Diamond Light Source",
    "license": "BSD-3-Clause",
    "author_email": "dials-support@lists.sourceforge.net",
    "project_urls": {
        "homepage": "https://dials.github.io",
        "repository": "https://github.com/cctbx/dxtbx",
    },
    "packages": setuptools.find_packages(where="src"),
    "package_dir": {"": "src"},
    "package_data": {
        "": ["*"],
        "dxtbx": ["array_family/*", "boost_python/*", "example/*", "py.typed"],
        "dxtbx.format": ["boost_python/*"],
        "dxtbx.masking": ["boost_python/*"],
        "dxtbx.model": ["boost_python/*"],
    },
    "entry_points": {
        "libtbx.precommit": ["dxtbx=dxtbx"],
        "libtbx.dispatcher.script": ["pytest=pytest"],
    },
}

build(setup_kwargs)
setuptools.setup(**setup_kwargs)



from __future__ import annotations

import sys

import libtbx
from scitbx.array_family import flex

from dxtbx.format.FormatCBFMiniEiger import FormatCBFMiniEiger
from dxtbx.masking import GoniometerMaskerFactory


class FormatCBFMiniEigerDLS16MSN160(FormatCBFMiniEiger):
    @staticmethod
    def understand(image_file):
        """Check to see if this format class can understand the image file.

        Args:
          image_file (str): The file path of the image file to check.

        Returns:
          bool: Returns ``True`` if the image_file is understood by this format class,
          else returns ``False``.

        """

        # this depends on DIALS for the goniometer shadow model; if missing
        # simply return False

        header = FormatCBFMiniEiger.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# detector" in record.lower()
                and "eiger" in record.lower()
                and "S/N 160-0001" in header
            ):
                return True

        return False

    @staticmethod
    def has_dynamic_shadowing(**kwargs):
        dynamic_shadowing = kwargs.get("dynamic_shadowing", False)
        if dynamic_shadowing in (libtbx.Auto, "Auto"):
            return True
        return dynamic_shadowing

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        self._dynamic_shadowing = self.has_dynamic_shadowing(**kwargs)
        super().__init__(image_file, **kwargs)

    def _goniometer(self):
        """Return a model for a multi-axis goniometer.

        This should be checked against the image header, though for miniCBF
        there are limited options for this.

        Returns:
          dxtbx.model.Goniometer.MultiAxisGoniometer: The goniometer model for
          this detector.

        """

        if "Phi" in self._cif_header_dictionary:
            phi = float(self._cif_header_dictionary["Phi"].split()[0])
        else:
            phi = 0.0

        if "Chi" in self._cif_header_dictionary:
            chi = float(self._cif_header_dictionary["Chi"].split()[0])
        else:
            chi = 0.0

        if "Omega" in self._cif_header_dictionary:
            omega = float(self._cif_header_dictionary["Omega"].split()[0])
        else:
            omega = 0.0

        phi_axis = (1, 0, 0)
        chi_axis = (0, 0, -1)
        omega_axis = (1, 0, 0)
        axes = flex.vec3_double((phi_axis, chi_axis, omega_axis))
        angles = flex.double((phi, chi, omega))
        names = flex.std_string(("GON_PHI", "GON_CHI", "GON_OMEGA"))
        return self._goniometer_factory.make_multi_axis_goniometer(
            axes, angles, names, scan_axis=2
        )

    def get_goniometer_shadow_masker(self, goniometer=None):
        if goniometer is None:
            goniometer = self.get_goniometer()

        assert goniometer is not None

        if goniometer.get_names()[1] == "GON_CHI":
            # SmarGon
            return GoniometerMaskerFactory.smargon(goniometer)

        else:
            raise ValueError(
                "Don't understand this goniometer: %s" % list(goniometer.get_names())
            )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniEigerDLS16MSN160.understand(arg))



from __future__ import annotations

import math

import h5py
import nxmx

from libtbx import Auto

from dxtbx.format.FormatNXmx import FormatNXmx
from dxtbx.format.FormatNXmxDLS import FormatNXmxDLS, get_bit_depth_from_meta
from dxtbx.masking import GoniometerMaskerFactory


class FormatNXmxDLSI19_2(FormatNXmxDLS):
    """
    Read NXmx-flavour data from beamline I19-2 at Diamond Light Source.

    Include the option of dynamic shadowing of the standard I19-2 diamond anvil
    pressure cell with a 76° conical aperture.
    """

    @staticmethod
    def understand(image_file):
        """This format class applies if the instrument name contains 'I19-2'."""
        with h5py.File(image_file, swmr=True) as handle:
            name = nxmx.h5str(FormatNXmx.get_instrument_name(handle))
        if name and "I19-2" in name:
            return True
        return False

    @staticmethod
    def has_dynamic_shadowing(**kwargs):
        """Check if dynamic shadowing should be applied for a diamond anvil cell."""
        dynamic_shadowing = kwargs.get("dynamic_shadowing", False)
        if dynamic_shadowing in (Auto, "Auto"):
            return False
        return dynamic_shadowing

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""
        self._dynamic_shadowing = self.has_dynamic_shadowing(**kwargs)
        super().__init__(image_file, **kwargs)
        try:
            self._bit_depth_readout = get_bit_depth_from_meta(self._meta)
        except Exception:
            self._bit_depth_readout = 16

    def get_goniometer_shadow_masker(self, goniometer=None):
        """Apply the dynamic mask for a diamond anvil cell with a 76° aperture."""
        if goniometer is None:
            goniometer = self.get_goniometer()

        return GoniometerMaskerFactory.diamond_anvil_cell(
            goniometer, cone_opening_angle=math.radians(76)
        )



"""An implementation of the SMV image reader for ADSC images"""


from __future__ import annotations

import calendar
import sys
import time

from iotbx.detectors import SMVImage

from dxtbx.format.FormatSMV import FormatSMV


class FormatSMVADSC(FormatSMV):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an ADSC SMV format image, i.e. we
        can make sense of it. Essentially that will be if it contains all of
        the keys we are looking for and not some we are not (i.e. that belong
        to a Rigaku Saturn.)"""

        size, header = FormatSMV.get_smv_header(image_file)

        # do not understand JHSim images
        if header.get("BEAMLINE") == "fake":
            return False

        # do not understand Timepix_SU images
        if header.get("BEAMLINE", "").upper() == "TIMEPIX_SU":
            return False

        # this used to include TIME
        wanted_header_items = [
            "BEAM_CENTER_X",
            "BEAM_CENTER_Y",
            "DISTANCE",
            "WAVELENGTH",
            "PIXEL_SIZE",
            "OSC_START",
            "OSC_RANGE",
            "SIZE1",
            "SIZE2",
            "BYTE_ORDER",
        ]

        if any(item not in header for item in wanted_header_items):
            return False

        unwanted_header_items = ["DTREK_DATE_TIME"]

        if any(item in header for item in unwanted_header_items):
            return False

        return True

    def detectorbase_start(self):
        if not hasattr(self, "detectorbase") or self.detectorbase is None:
            self.detectorbase = SMVImage(self._image_file)
            self.detectorbase.open_file = self.open_file
            self.detectorbase.readHeader()

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. Invert the axis if
        the rotation range is negative."""

        if float(self._header_dictionary["OSC_RANGE"]) > 0:
            return self._goniometer_factory.single_axis()
        else:
            return self._goniometer_factory.single_axis_reverse()

    def _adsc_trusted_range(self):
        """The data is 16-bit uint, where the upper value, 65535, is untrusted
        as overloads will be truncated to this value.

        ADSC CCD detectors always have a software pedestal, even if the value is
        not recorded in the header. In practice, this means that useful data
        values are always > 0, while the pixels at module boundaries are == 0"""

        max_trusted_value = 65535 - 1
        min_trusted_value = 1

        return min_trusted_value, max_trusted_value

    def _adsc_module_gain(self, model=None):
        """Return an appropriate gain value in ADU per captured X-ray for an
        ADSC CCD module. If the model is None (unknown) then make a guess based
        on header details"""

        # Values are based on a combination of manufacturer datasheets,
        # James Holton's list at http://bl831.als.lbl.gov/xtalsize.html and
        # empirical guesswork based on spot finding results. The accuracy should
        # not be expected to be better than 20% and indeed may be worse than that.
        # Values may refer to gain in ADU per incident photon rather than the
        # more appropriate ADU per captured photon.

        # Allow gain set in the header to override the guesswork here
        if "GAIN" in self._header_dictionary:
            return float(self._header_dictionary["GAIN"])

        # Get the binning level
        bin_lev = str(self._header_dictionary.get("BIN"))

        if model is None:
            # guesswork based on the header details
            npx1 = int(self._header_dictionary["SIZE1"])
            npx2 = int(self._header_dictionary["SIZE2"])
            px_sz = float(self._header_dictionary["PIXEL_SIZE"])
            if npx1 == npx2 == 2304 and px_sz == 0.0816:
                model = "Q4"  # or Q4R
            elif npx1 == npx2 == 4096 and px_sz == 0.0512:
                model = "Q210"  # or Q210R bin none or 1x1
            elif npx1 == npx2 == 2048 and px_sz == 0.1024:
                model = "Q210"  # or Q210R bin 2x2
                bin_lev = "2x2"
            elif npx1 == npx2 == 6144 and px_sz == 0.0512:
                model = "Q315"  # or Q315R bin none or 1x1
            elif npx1 == npx2 == 3072 and px_sz == 0.1024:
                model = "Q315"  # or Q315R bin 2x2
                bin_lev = "2x2"
            elif npx1 == npx2 == 4168 and px_sz == 0.0648:
                model = "Q270"  # or Q270R bin none or 1x1
            elif npx1 == npx2 == 2084 and px_sz == 0.0324:
                model = "Q270"  # or Q270R bin 2x2
                bin_lev = "2x2"
            else:
                # Unidentified model
                return 1.0

        model = model.upper()
        if model.startswith("Q270"):
            # Don't know the difference between binning modes in this case
            return 2.8

        if model.startswith("Q4"):
            # https://web.archive.org/web/20051224172252/http://www.adsc-xray.com:80/prod_compare.html
            # Don't know the difference between binning modes in this case
            return 1.25

        # Get binning mode HW/SW
        bin_type = self._header_dictionary.get("BIN_TYPE")

        if bin_lev.upper() == "NONE":
            bin_type = None
        elif bin_type != "HW":
            bin_type = "SW"

        if bin_type == "SW":
            # return 0.6 This does not look believable. Default to 2.4 instead
            return 2.4

        if bin_type != "HW":  # assume unbinned
            return 2.4
        elif model.endswith("R"):  # e.g. Q315r, HW binning
            return 1.8
        else:  # e.g. Q315, HW binning
            return 2.4

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._header_dictionary["DISTANCE"])
        beam_x = float(self._header_dictionary["BEAM_CENTER_X"])
        beam_y = float(self._header_dictionary["BEAM_CENTER_Y"])
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        image_size = (
            float(self._header_dictionary["SIZE1"]),
            float(self._header_dictionary["SIZE2"]),
        )

        return self._detector_factory.simple(
            "CCD",
            distance,
            (beam_y, beam_x),
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            self._adsc_trusted_range(),
            [],
            gain=self._adsc_module_gain(),
            pedestal=float(self._header_dictionary.get("IMAGE_PEDESTAL", 0)),
        )

    def _beam(self):
        """Return a simple model for the beam."""

        wavelength = float(self._header_dictionary["WAVELENGTH"])

        return self._beam_factory.simple(wavelength)

    def _scan(self):
        """Return the scan information for this image."""
        exposure_time = float(self._header_dictionary["TIME"])
        epoch = None

        # PST, PDT timezones not recognised by default...

        epoch = 0
        try:
            date_str = self._header_dictionary["DATE"]
            date_str = date_str.replace("PST", "").replace("PDT", "")
        except KeyError:
            date_str = ""
        for format_string in ["%a %b %d %H:%M:%S %Y", "%a %b %d %H:%M:%S %Z %Y"]:
            try:
                epoch = calendar.timegm(time.strptime(date_str, format_string))
                break
            except ValueError:
                pass

        # assert(epoch)
        osc_start = float(self._header_dictionary["OSC_START"])
        osc_range = abs(float(self._header_dictionary["OSC_RANGE"]))

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        assert len(self.get_detector()) == 1
        panel = self.get_detector()[0]
        image_size = panel.get_image_size()
        return self._get_endianic_raw_data(size=image_size)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVADSC.understand(arg))



"""
An implementation of the SMV image reader for ADSC images. Inherits from
FormatSMVADSC, customised for example on Australian Synchrotron SN 457
which has reversed phi.
"""


from __future__ import annotations

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSNSN457(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument number 457."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 457."""

        # check this is detector serial number 457

        size, header = FormatSMVADSCSN.get_smv_header(image_file)

        return int(header["DETECTOR_SN"]) == 457

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header."""

        return self._goniometer_factory.single_axis_reverse()



"""Implementation of an ImageFormat class to read MarIP-format image"""


from __future__ import annotations

import sys

from iotbx.detectors import EDFWrapper

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format


class FormatEDFALS733(Format):
    """EDF is the ESRF data format.
    In present context, ALS 7.3.3 uses this format for its single-tile Pilatus
    """

    @staticmethod
    def understand(image_file):
        try:
            with FormatEDFALS733.open_file(image_file, "rb") as fh:
                tag = fh.read(10)
        except OSError:
            return False

        return tag == b"{\nHeaderID"

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        super().__init__(image_file, **kwargs)

    def detectorbase_start(self):
        pass

    def _start(self):
        """Open the image file, read the image header, copy the key / value
        pairs into an internal dictionary self._header_dictionary along with
        the length of the header in bytes self._header_size."""
        self.detectorbase = EDFWrapper(self._image_file)
        self.detectorbase.readHeader()

    def _goniometer(self):
        return self._goniometer_factory.single_axis()

    def _detector(self):
        """Return a model for a simple detector"""
        return self._detector_factory.simple(
            sensor="PAD",
            distance=self.detectorbase.parameters["DISTANCE"],
            beam_centre=(
                self.detectorbase.parameters["BEAM_CENTER_X"],
                self.detectorbase.parameters["BEAM_CENTER_Y"],
            ),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(
                self.detectorbase.parameters["PIXEL_SIZE"],
                self.detectorbase.parameters["PIXEL_SIZE"],
            ),
            image_size=(
                self.detectorbase.parameters["SIZE1"],
                self.detectorbase.parameters["SIZE2"],
            ),
            trusted_range=(0, self.detectorbase.saturation),
            mask=[],
        )  # a list of dead rectangles

    def _beam(self):
        """Return a simple model for the beam."""
        return self._beam_factory.simple(wavelength=1.0)  # dummy argument

    def _scan(self):
        """Return the scan information for this image."""
        return self._scan_factory.single_file(
            filename=self._image_file,
            exposure_times=self.detectorbase.parameters["count_time"],
            osc_start=0.0,
            osc_width=0.0,
            epoch=None,
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatEDFALS733.understand(arg))



"""
An implementation of the SMV image reader for Rigaku images.
Inherits from FormatSMV.
"""

from __future__ import annotations

import calendar
import sys
import time

from dxtbx.format.FormatSMV import FormatSMV


class FormatSMVRigaku(FormatSMV):
    """A class for reading SMV format Rigaku images. 'Abstract' class."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like a Rigaku d*TREK SMV format image,
        i.e. we can make sense of it. Essentially that will be if it contains
        all of the keys we are looking for."""

        size, header = FormatSMV.get_smv_header(image_file)

        wanted_header_items = [
            "DETECTOR_NUMBER",
            "DETECTOR_NAMES",
            "BYTE_ORDER",
            "DIM",
            "SIZE1",
            "SIZE2",
            "Data_type",
        ]

        if any(header_item not in header for header_item in wanted_header_items):
            return False

        # code around CMOS1 on ALS 4.2.2

        detector_prefixes = header["DETECTOR_NAMES"].split()

        if len(detector_prefixes) == 1:
            prefix = detector_prefixes[0]
            det_desc = "%sDETECTOR_DESCRIPTION" % prefix
            if "CMOS-1" in header.get(det_desc, ""):
                return False

        detector_prefixes = header["DETECTOR_NAMES"].split()
        try:
            detector_number = int(header["DETECTOR_NUMBER"].strip())
        except (KeyError, AttributeError, ValueError):
            return False

        if detector_number != len(detector_prefixes):
            return False

        return True

    def _goniometer(self):
        """Overload this method to read the image file however you like so
        long as the result is an goniometer."""

        raise NotImplementedError("overload me")

    def _detector(self):
        """Overload this method to read the image file however you like so
        long as the result is an detector."""

        raise NotImplementedError("overload me")

    def _beam(self):
        """Overload this method to read the image file however you like so
        long as the result is an beam."""

        raise NotImplementedError("overload me")

    def _scan(self):
        """Overload this method to read the image file however you like so
        long as the result is an scan."""

        raise NotImplementedError("overload me")

    def _create_single_SVM_scan(self, epoch_time_struct, local_time=True):
        """Return the scan information for this image."""

        rotation = self.get_rotation()

        if local_time:
            epoch = time.mktime(epoch_time_struct)
        else:
            epoch = calendar.timegm(epoch_time_struct)

        osc_start = rotation[0]
        osc_range = rotation[2]
        exposure_time = rotation[3]

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        assert len(self.get_detector()) == 1
        panel = self.get_detector()[0]
        image_size = panel.get_image_size()
        return self._get_endianic_raw_data(size=image_size)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVRigaku.understand(arg))



from __future__ import annotations

from cctbx.sgtbx import space_group as SG
from scitbx import matrix

try:
    from ..dxtbx_model_ext import (
        Crystal,
        MosaicCrystalKabsch2010,
        MosaicCrystalSauter2014,
    )
except ModuleNotFoundError:
    from dxtbx_model_ext import (  # type: ignore
        Crystal,
        MosaicCrystalKabsch2010,
        MosaicCrystalSauter2014,
    )


class CrystalFactory:
    @staticmethod
    def from_dict(d, t=None):
        """Convert the dictionary to a crystal model

        Params:
            d The dictionary of parameters
            t The template dictionary to use

        Returns:
            The crystal model

        """
        if d is None and t is None:
            return None
        joint = t.copy() if t else {}
        joint.update(d)

        # Create the model from the dictionary
        if "ML_half_mosaicity_deg" in joint:
            assert "ML_domain_size_ang" in joint
            if (
                joint["ML_half_mosaicity_deg"] is None
                or joint["ML_domain_size_ang"] is None
            ):
                assert (
                    joint["ML_half_mosaicity_deg"] is None
                    and joint["ML_domain_size_ang"] is None
                )
            else:
                if joint.get("mosaicity", 0) > 0:
                    print(
                        "Warning, two kinds of mosaicity found. Using Sauter2014 model"
                    )
                return MosaicCrystalSauter2014.from_dict(joint)
        if "mosaicity" in joint:
            return MosaicCrystalKabsch2010.from_dict(joint)
        else:
            return Crystal.from_dict(joint)

    @staticmethod
    def from_mosflm_matrix(
        mosflm_A_matrix, unit_cell=None, wavelength=None, space_group=None
    ):
        """
        Create a crystal_model from a Mosflm A matrix (a*, b*, c*); N.B. assumes
        the mosflm coordinate frame::

                                                         /!
                            Y-axis                      / !
                              ^                        /  !
                              !                       /   !
                              !                      /    !
                              !   /                 /  Xd !
                              !  /                 / * ^  !
                              ! /                  ! 3 !  !
                              !/      X-ray beam   !   !  !
                              /------------------------/--!---->X-axis
                             /                     !  / *1!
                          <-/-                     ! /    !
                           / \\+ve phi             ! Yd  /
                          /   /                    ! 2  /
                         /                         ! * /
                        Z-axis                  Ys ^ _/
                      Rotation                     ! /| Xs
                       axis                        !/
                                                   O

        Also assume that the mosaic spread is 0.

        :param mosflm_A_matrix: The A matrix in Mosflm convention.
        :type mosflm_A_matrix: tuple of floats
        :param unit_cell: The unit cell parameters which are used to determine the
                          wavelength from the Mosflm A matrix.
        :type unit_cell: cctbx.uctbx.unit_cell
        :param wavelength: The wavelength to scale the A matrix
        :type wavelength: float
        :param space_group: If the space group is None then the space_group will
                            be assigned as P1
        :type space_group: cctbx.sgtbx.space_group
        :returns: A crystal model derived from the given Mosflm A matrix
        :rtype: :py:class:`crystal_model`
        """
        if not space_group:
            space_group = SG("P1")

        A_star = matrix.sqr(mosflm_A_matrix)
        A = A_star.inverse()

        if unit_cell:
            unit_cell_constants = unit_cell.parameters()
            a = matrix.col(A.elems[0:3])
            wavelength = unit_cell_constants[0] / a.length()
            A *= wavelength
        elif wavelength:
            A *= wavelength
        else:
            # assume user has pre-scaled
            pass

        a = A.elems[0:3]
        b = A.elems[3:6]
        c = A.elems[6:9]
        rotate_mosflm_to_imgCIF = matrix.sqr((0, 0, 1, 0, 1, 0, -1, 0, 0))
        _a = rotate_mosflm_to_imgCIF * a
        _b = rotate_mosflm_to_imgCIF * b
        _c = rotate_mosflm_to_imgCIF * c

        return Crystal(_a, _b, _c, space_group=space_group)



from __future__ import annotations

import math
import os
import pickle
import random

import pytest

from cctbx.eltbx import attenuation_coefficient
from libtbx.test_utils import approx_equal
from scitbx import matrix
from scitbx.array_family import flex

import dxtbx
from dxtbx.model import (
    OffsetParallaxCorrectedPxMmStrategy,
    OffsetPxMmStrategy,
    Panel,
    ParallaxCorrectedPxMmStrategy,
    SimplePxMmStrategy,
)
from dxtbx.model.beam import BeamFactory
from dxtbx.model.detector import DetectorFactory


@pytest.fixture
def model(dials_regression):
    filename = os.path.join(dials_regression, "image_examples", "XDS", "XPARM.XDS")

    models = dxtbx.load(filename)
    detector = models.get_detector()
    assert len(detector) == 1
    detector = detector[0]
    t0 = 0.320

    table = attenuation_coefficient.get_table("Si")
    mu = table.mu_at_angstrom(models.get_beam().get_wavelength()) / 10.0
    pixel_size = detector.get_pixel_size()

    return {"mu": mu, "t0": t0, "detector": detector, "pixel_size": pixel_size}


def correct_gold(model, xy):
    mu = model["mu"]
    t0 = model["t0"]
    s1 = matrix.col(model["detector"].get_lab_coord(xy)).normalize()
    d1 = matrix.col(model["detector"].get_fast_axis())
    d2 = matrix.col(model["detector"].get_slow_axis())
    dn = d1.cross(d2)
    cos_theta = s1.dot(dn)
    t = t0 / cos_theta
    o = (1.0 / mu) - (t + 1.0 / mu) * math.exp(-mu * t)
    cx = xy[0] + s1.dot(d1) * o
    cy = xy[1] + s1.dot(d2) * o
    return (cx / model["pixel_size"][0], cy / model["pixel_size"][1])


def test_correction_on_random_coordinates(model):
    convert = ParallaxCorrectedPxMmStrategy(model["mu"], model["t0"])
    for i in range(10000):
        xy = matrix.col((random.uniform(-1000, 1000), random.uniform(-1000, 1000)))

        # Do the forward and reverse corrections
        xy_corr_gold = matrix.col(correct_gold(model, xy))
        xy_corr = matrix.col(convert.to_pixel(model["detector"], xy))
        xy_corr_inv = matrix.col(convert.to_millimeter(model["detector"], xy_corr))

        # Check the values
        assert abs(xy_corr_gold - xy_corr) < 1e-7
        assert abs(xy_corr_inv - xy) < 1e-3


def test_array(model):
    xy = flex.vec2_double(
        (random.uniform(-1000, 1000), random.uniform(-1000, 1000)) for i in range(100)
    )
    xy_corr = model["detector"].get_lab_coord(xy)
    xy_corr_panel = model["detector"].get_lab_coord(xy)
    xy_corr_gold = [model["detector"].get_lab_coord(xy_single) for xy_single in xy]
    assert approx_equal(xy_corr, xy_corr_gold)
    assert approx_equal(xy_corr_panel, xy_corr_gold)


def test_inverted_axis():
    def get_values(invert_y):
        beam = BeamFactory.simple(wavelength=1)

        if invert_y:
            y_direction = "-y"
        else:
            y_direction = "+y"

        detector = DetectorFactory.simple(
            sensor=DetectorFactory.sensor("PAD"),
            distance=100,
            beam_centre=[50, 50],
            fast_direction="+x",
            slow_direction=y_direction,
            pixel_size=[0.1, 0.1],
            image_size=[1000, 1000],
        )[0]

        wavelength = beam.get_wavelength()
        thickness = 0.5
        table = attenuation_coefficient.get_table("Si")
        mu = table.mu_at_angstrom(wavelength) / 10.0
        t0 = thickness

        for panel in detector:
            panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
        v1 = detector.pixel_to_millimeter((0, 0))
        v2 = detector.pixel_to_millimeter((1000, 1000))

        return v1, v2

    v11, v12 = get_values(False)
    v21, v22 = get_values(False)

    assert abs(v11[0] - v21[0]) < 1e-7
    assert abs(v11[1] - v21[1]) < 1e-7
    assert abs(v12[0] - v22[0]) < 1e-7
    assert abs(v12[1] - v22[1]) < 1e-7


def test_offset_px_mm_strategy():
    # for future reference this is the array the same shape
    # as the image in pixels with offsets in pixels

    dx = flex.double(flex.grid(10, 10), 1)
    dy = flex.double(flex.grid(10, 10), 1)

    strategy = OffsetParallaxCorrectedPxMmStrategy(1, 1, dx, dy)

    p = Panel()
    p.set_image_size((10, 10))
    p.set_mu(1)
    p.set_thickness(1)
    p.set_px_mm_strategy(strategy)

    d = p.to_dict()

    # Panel.from_dict sets the strategy to ParallaxCorrectedPxMmStrategy rather than
    # OffsetParallaxCorrectedPxMmStrategy as the offsets aren't currently serialized
    pnew = Panel.from_dict(d)
    assert pnew != p

    pnew = pickle.loads(pickle.dumps(pnew))
    assert pnew != p


def test_px_mm_strategy_equality():
    simple = SimplePxMmStrategy()
    assert simple == SimplePxMmStrategy()

    parallax_corrected = ParallaxCorrectedPxMmStrategy(1, 1)
    assert parallax_corrected == ParallaxCorrectedPxMmStrategy(1, 1)
    assert parallax_corrected != ParallaxCorrectedPxMmStrategy(1, 2)
    assert parallax_corrected != ParallaxCorrectedPxMmStrategy(2, 1)
    assert parallax_corrected != simple

    dx = flex.double(flex.grid(10, 10), 1)
    dy = flex.double(flex.grid(10, 10), 1)

    offset = OffsetPxMmStrategy(dx, dy)
    assert offset == OffsetPxMmStrategy(dx, dy)
    assert offset != simple
    assert offset != parallax_corrected

    offset_parallax_corrected = OffsetParallaxCorrectedPxMmStrategy(1, 1, dx, dy)
    assert offset_parallax_corrected == OffsetParallaxCorrectedPxMmStrategy(
        1, 1, dx, dy
    )
    assert offset_parallax_corrected != OffsetParallaxCorrectedPxMmStrategy(
        1, 2, dx, dy
    )
    assert offset_parallax_corrected != OffsetParallaxCorrectedPxMmStrategy(
        2, 1, dx, dy
    )
    assert offset_parallax_corrected != simple
    assert offset_parallax_corrected != parallax_corrected
    assert offset_parallax_corrected != offset



from __future__ import annotations

import argparse

import dxtbx.util
from dxtbx.format import Registry

dag = Registry.get_format_class_dag()


def print_class(class_name: str, filename: str | None = None, depth=1):
    """Print a Format class name if it matches a file"""
    if filename is None or (
        Registry.get_format_class_for(class_name).understand(filename)
    ):
        print(f"{depth: 5} {'  ' * depth} {class_name}")
    else:
        return

    for child in dag.get(class_name, []):
        print_class(child, filename, depth + 1)


def show_registry(filename: str | None = None):
    if filename:
        print(f"Format classes that understand {filename}:")
    else:
        print("Format classes in the dxtbx registry:")

    print("Depth  Class name")
    print("    0  Format")

    for class_name in dag["Format"]:
        print_class(class_name, filename)


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser(
        description="Show hierarchy of dxtbx format classes"
    )
    parser.add_argument(
        "filename", help="Only show classes that understand this file", nargs="?"
    )
    opts = parser.parse_args(args)
    show_registry(opts.filename)


if __name__ == "__main__":
    run()



from __future__ import annotations

import pytest

from dxtbx.format.FormatCBFFullPilatusDLS6MSN126 import FormatCBFFullPilatusDLS6MSN126
from dxtbx.masking import SmarGonShadowMasker
from dxtbx.model.experiment_list import ExperimentListFactory


def test_DLS_I03_smargon(dials_data):
    filename = (
        dials_data("image_examples", pathlib=True) / "DLS_I03_smargon_0001.cbf.gz"
    )
    assert FormatCBFFullPilatusDLS6MSN126.understand(filename)
    expts = ExperimentListFactory.from_filenames(
        [filename], format_kwargs={"dynamic_shadowing": True}
    )
    assert len(expts) == 1
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatCBFFullPilatusDLS6MSN126
    detector = imageset.get_detector()
    assert detector[0].get_trusted_range() == (0, 91794)
    gonio = imageset.get_goniometer()
    assert list(gonio.get_angles()) == pytest.approx([45.0, 45.0, 45.0])
    assert list(gonio.get_axes().as_double()) == pytest.approx(
        [1.0, 0.0, 0.0, 0.0, 0.0, -1.0, 1.0, 0.0, 0.0]
    )
    assert list(gonio.get_names()) == ["GON_PHI", "GON_CHI", "GON_OMEGA"]
    assert imageset.has_dynamic_mask()
    masker = imageset.masker()
    assert isinstance(masker, SmarGonShadowMasker)
    assert masker.get_mask(imageset.get_detector(), 0)[0].count(False) == 0
    assert masker.get_mask(imageset.get_detector(), 100)[0].count(False) == 261588



from __future__ import annotations

import sys

from iotbx.detectors.raxis_nonsquare import NonSquareRAXISImage

from dxtbx.format.Format import Format
from dxtbx.format.FormatRAXIS import RAXISHelper


class FormatRAXISII(RAXISHelper, Format):
    @staticmethod
    def understand(image_file):
        try:
            with FormatRAXISII.open_file(image_file, "rb") as fh:
                return fh.read(7) == b"R-AXIS2"
        except OSError:
            return False

    def detectorbase_start(self):
        pass  # override with an empty function

    def _start(self):
        self.detectorbase = NonSquareRAXISImage(self._image_file)
        self.detectorbase.readHeader()

    def _goniometer(self):
        return self._goniometer_factory.single_axis()

    def _detector(self):
        """Return a model for a simple detector"""

        return self._detector_factory.simple(
            sensor="IMAGE_PLATE",
            distance=self.detectorbase.distance,
            beam_centre=(self.detectorbase.beamx, self.detectorbase.beamy),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(self.detectorbase.pixel_size, self.detectorbase.pixel_size),
            image_size=(self.detectorbase.size1, self.detectorbase.size2),
            trusted_range=(0, self.detectorbase.saturation),
            mask=[],
        )  # a list of dead rectangles

    def _beam(self):
        """Return a simple model for the beam."""

        return self._beam_factory.simple(self.detectorbase.wavelength)

    def _scan(self):
        """Return the scan information for this image."""

        return self._scan_factory.single_file(
            filename=self._image_file,
            exposure_times=1,
            osc_start=self.detectorbase.parameters["OSC_START"],
            osc_width=self.detectorbase.parameters["OSC_RANGE"],
            epoch=None,
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatRAXISII.understand(arg))



from __future__ import annotations


def id_conversion(path):
    return path.replace("/", "-")


smv_images = [
    "image_examples/ALS_501/als501_q4_1_001.img",
    "image_examples/SPring8_BL26B1_SaturnA200/A200_000001.img",
    "image_examples/SPring8_BL26B1_SaturnA200/A200_000002.img",
    "image_examples/APS_19ID/q315_unbinned_a.0001.img",
    "image_examples/ALS_821/q210_lyso_1_101.img",
    "image_examples/MLFSOM_simulation/fake_00001.img",
    "image_examples/ALS_831/q315r_lyso_001.img",
    "image_examples/DESY_ID141/q210_2_001.img",
    "image_examples/SSRL_bl91/q315_1_001.img",
    "image_examples/APS_24IDC/q315_1_001.img",
    "image_examples/ALS_1231/q315r_lyso_1_001.img",
    "image_examples/SRS_142/q4_1_001.img",
    "image_examples/ALS_422/lyso_041013a_1_001.img",
    "image_examples/APS_17ID/q210_1_001.img",
    "image_examples/saturn/lyso_00001.img",
]
smv_image_ids = list(map(id_conversion, smv_images))

tiff_images = [
    "image_examples/SPring8_BL32XU_MX225HS/ds_000045.img",
    "image_examples/SPring8_BL32XU_MX225HS/ds_000001.img",
    "image_examples/SPring8_BL44XU_MX300HE/bl44xu_lys_000002.img",
    "image_examples/SPring8_BL44XU_MX300HE/bl44xu_lys_000001.img",
    "image_examples/SPring8_BL32XU/rayonix225hs_0001.img",
    "image_examples/SPring8_BL32XU/rayonix225_0001.img",
    "image_examples/SLS_X06SA/mar225_2_001.img",
    "image_examples/CLS1_08ID1/mar225_2_E0_0001.img",
    "image_examples/SPring8_BL38B1_MX225HE/bl38b1_001.img",
    "image_examples/SPring8_BL38B1_MX225HE/bl38b1_090.img",
    "image_examples/SRS_101/mar225_001.img",
    "image_examples/SPring8_BL12B2_MX225HE/lys001_000091.img",
    "image_examples/SPring8_BL12B2_MX225HE/lys001_000001.img",
    "image_examples/SPring8_BL26B2_MX225/2sec_Al200um_000001.img",
    "image_examples/SPring8_BL26B2_MX225/2sec_Al200um_000090.img",
]
tiff_image_ids = list(map(id_conversion, tiff_images))

cbf_images = [
    "image_examples/ESRF_ID29/trypsin_1_0001.cbf",
    "image_examples/xia2/merge2cbf_averaged_0001.cbf",
    "image_examples/dials-190/whatev1_01_00002.cbf",
    "image_examples/dials-190/whatev1_02_00001.cbf",
    "image_examples/dials-190/whatev1_02_00002.cbf",
    "image_examples/dials-190/whatev1_03_00001.cbf",
    "image_examples/dials-190/whatev1_01_00001.cbf",
    "image_examples/dials-190/whatev1_03_00002.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0005.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0004.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0006.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0008.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0009.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0010.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0007.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0002.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0001.cbf",
    "image_examples/APS_24IDE_test/thaum-12_1_0003.cbf",
    "image_examples/APS_24IDC/pilatus_1_0001.cbf",
    "image_examples/SLS_Eiger_16M_as_CBF/insu_with_bs_labelit_0901.cbf",
    "image_examples/SLS_Eiger_16M_as_CBF/insu_with_bs_labelit_0001.cbf",
    "image_examples/SPring8_BL41XU_PILATUS3_6M/data1_000001.cbf",
    "image_examples/SPring8_BL41XU_PILATUS3_6M/data1_000901.cbf",
    "image_examples/DLS_I02/X4_wide_M1S4_1_0001.cbf",
    "image_examples/DLS_I23/I23_P12M_alpha_0001.cbf",
    "image_examples/DLS_I23/germ_13KeV_0001.cbf",
    "image_examples/SLS_X06SA/pilatus6m_1_00001.cbf",
    "image_examples/SPring8_ADSC_SN916/Xtal17-2phi_3_015.cbf",
    "image_examples/DLS_I19/I19_P300k_00001.cbf",
    "image_examples/ED_From_TIFF/170112330001.cbf",
]
cbf_image_ids = list(map(id_conversion, cbf_images))

cbf_multitile_images = [
    "stills_test_data/hit-20111202210224984.cbf",
    "stills_test_data/hit-s00-20140306002935980.cbf",
    "stills_test_data/hit-s00-20140306002857363.cbf",
]
cbf_multitile_image_ids = list(map(id_conversion, cbf_multitile_images))

hdf5_images = ["image_examples/putative_imgCIF_HDF5_mapping/minicbf.h5"]
hdf5_image_ids = list(map(id_conversion, hdf5_images))
